//==============================================================================
//Copyright(c), Meilhaus Electronic GmbH
//              http://www.meilhaus.de
//              support@meilhaus.de
//
//Source File : uMain.pas
//Description : Example program for reading values from Frequency-In (FI) subdevices
//              - Open main driver (ME-iDS)
//              - Go through the list of all devices and find FI subdevices
//              - Add FI subdevices to the combobox, so the user can select one
//              - Read data from selected subdevice when user clicks on button "Read Data"
//              - Close main driver (ME-iDS) when program is closed
//==============================================================================
//  Note about the hardware:
//  After meIOSingle is called (start of the measurement operation)
//  the FI hardware will automatically detect the next rising edge
//  of the unknown digital signal to start the measurement of the signal period.
//  The first counter (counter A) on the hardware is started.
//  Counter A is stopped when the next falling edge is detected.
//  At the same time a second counter (counter B) is started that will stop on
//  detection of the next rising edge (the end of the signal period).
//  The two counter values describe the signal period
//  The ME-iDS function library will provide two values
//  1. the counter value for the entire signal period (counter A + counter B)
//  2. the counter value of the 1st phase of the signal period (counter A)
//
//  Reading Data:
//  A new measurement operation is started with the given timeout value
//  (iFlag is set to ME_IO_SINGLE_CONFIG_NO_FLAGS)
//  After the measuring is done (meIOSingle() returned successfully)
//  the value of Total ticks is returned in member iValue of meIOSingle_t,
//  Then meIOSingle is called a second time to read the value of the first phase.
//  Therefore ME_IO_SINGLE_TYPE_FI_LAST_VALUE must be specified in iFlags
//  (If ME_IO_SINGLE_TYPE_FI_LAST_VALUE isn't specified a new measuring
//  would be started and the returned counter value wouldn't belong
//  to the same signal period, thus giving wrong results)
//  With the function meIOSingleTicksToTime() these values can be converted to a time value.
//
//  unknown digital signal:
//      +-------+                   +---
//      |       |                   |
//      |       |                   |
//  ----+       +-------------------+
//
//  ^---^-------^-------------------^
//  0   1       2                   3
//
//  0: Start of the operation (call of function meIOSingle())
//  1: Start of hardware counter A on rising edge of the signal
//  2: End of the first phase of the signal, stop of hardware counter A
//     Start of hardware counter B
//  3: End of the second phase of the signal, stop of hardware counter B
//     End of the operation (function meIOSingle() returned)
//
//------------------------------------------------------------------------------
unit uMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, ComCtrls, Spin, meIDS, meError;

type
//-------------------------------------------------
{TFISubDevices}
  TFiSubdevice = class(TCollectionItem)
  private
    FDeviceName: string;
    FIndexDevice: Integer;
    FIndexSubdevice: Integer;
    FNumberOfChannels: Integer;
  published
    property DeviceName: string read FDeviceName;
    property IndexDevice: Integer read FIndexDevice;
    property IndexSubdevice: Integer read FIndexSubdevice;
    property NumberOfChannels: Integer read FNumberOfChannels;
  end;

  TFiSubdevices = class(TCollection)
  private
    function GetSubdevice(Index: Integer): TFiSubdevice;
  public
    function AddSubdevice(ADeviceName: string; AIdxDevice: Integer; AIdxSubdevice: Integer; ANumOfChannels: Integer): TFiSubdevice;
    property Items[Index: Integer]: TFiSubdevice read GetSubdevice;
    Constructor Create;
  end;
//--End--------------------------------------------

  TfrmMain = class(TForm)
    btnClearHwInfo: TButton;
    btnAbout: TButton;
    grpBox1: TGroupBox;
    lbl3: TLabel;
    coBoxSubdevs: TComboBox;
    lbl1: TLabel;
    memoInfo: TMemo;
    btnClose: TButton;
    btnDeviceRead: TButton;
    btnReset: TButton;
    lbl2: TLabel;
    editTimeout: TEdit;
    udTimeout: TUpDown;
    lbl4: TLabel;
    coBoxChannel: TComboBox;
    procedure btnAboutClick(Sender: TObject);
    procedure btnCloseClick(Sender: TObject);
    procedure btnClearHwInfoClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure btnResetClick(Sender: TObject);
    procedure btnDeviceReadClick(Sender: TObject);
    procedure coBoxSubdevsChange(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    { Private-Deklarationen }
    openErrSuccess: Integer;
    lineNumber: Integer;
    fiSubdevices: TFiSubdevices;
    function MeVersionToString(version: Integer): string;
    function GetSubDevTypeName(subdeviceType: Integer): string;
    procedure GetFiSubdevices(numDevices: Integer);
    procedure GetSubdevicesByType(idxDevice: Integer; subdeviceType: Integer);
    procedure OpenMEiDSDriver();
    procedure CloseMEiDSDriver();
    procedure DisplayVersionLibrary();
    procedure DisplayVersionMainDriver();
    procedure DisplayErrorMessage(ErrMsg: string; meError: Integer);
    function GetDevicePlugged(idxDevice: Integer; var plugged: Integer): Integer;
    procedure DisplaySubdevices();
    procedure AddFiSubdevice(devName: string; idxDevice: Integer; idxSubdevice: Integer; subdeviceType: Integer);
    function ReadSingleValue(fiSubdevice: TFiSubdevice; idxChannel: Integer;
      timeOut: Integer; flags: Integer; var value: Integer): Integer;
    function ConfigureSubdeviceForInput(fiSubdevice: TFiSubdevice; idxChannel: Integer): Integer;
    function CalculateTime(fiSubdevice: TFiSubdevice; timer: Integer;
      valueDigital: Integer; var time: Double): Integer;
    function DisplayValues(fiSubdevice: TFiSubdevice; idxChannel: Integer;
      valueTotal: Integer; value1stPhase: Integer): Integer;
  public
    { Public-Deklarationen }
  end;

var
  frmMain: TfrmMain;

implementation

uses uAbout;

{$R *.dfm}

const
  ME_IDS_NOTOPENED = -1;

//----------------------------------------------
{TFiSubDevices}
Constructor TFiSubdevices.Create();
begin
  inherited Create(TFiSubdevice);
end;

function TFiSubdevices.GetSubdevice(Index: Integer): TFiSubdevice;
begin
  Result := TFiSubdevice(inherited GetItem(Index));
end;

function TFiSubdevices.AddSubdevice(ADeviceName: string; AIdxDevice: Integer; AIdxSubdevice: Integer; ANumOfChannels: Integer): TFiSubdevice;begin
  Result := inherited Add as TFiSubdevice;
  Result.FDeviceName := ADeviceName;
  Result.FIndexDevice := AIdxDevice;
  Result.FIndexSubdevice := AIdxSubdevice;
  Result.FNumberOfChannels := ANumOfChannels;
end;

//----------------------------------------------
{TfrmMain}
procedure TfrmMain.FormCreate(Sender: TObject);
begin
  openErrSuccess := ME_IDS_NOTOPENED;
  lineNumber := 0;
  fiSubdevices := TFiSubdevices.Create;
end;

procedure TfrmMain.FormShow(Sender: TObject);
begin
  OpenMEiDSDriver();
end;

procedure TfrmMain.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  CloseMEiDSDriver();
  fiSubdevices.Free;
end;

procedure TfrmMain.OpenMEiDSDriver();
var
  numDevices: Integer;
begin
  memoInfo.Lines.Append('Opening Main driver...');
  openErrSuccess := meOpen(ME_OPEN_NO_FLAGS);

  if openErrSuccess = ME_ERRNO_SUCCESS then
  begin
    DisplayVersionLibrary;
    DisplayVersionMainDriver;
    memoInfo.Lines.Append('');

    meQueryNumberDevices(numDevices);
    memoInfo.Lines.Append('Number of devices found: ' + IntToStr(numDevices));

    GetFiSubdevices(numDevices);
  end else
  begin
    DisplayErrorMessage('Error opening Main driver. Error: ', openErrSuccess);   
  end;
end;

procedure TfrmMain.GetFiSubdevices(numDevices: Integer);
var
  meError: Integer;
  idxDevice: Integer;
  plugged: Integer;
begin
  if (openErrSuccess = ME_ERRNO_SUCCESS) and (numDevices > 0) then
  begin
    memoInfo.Lines.Add('Retrieving sub-devices of type ME_TYPE_FI (Frequency In)...');
    for idxDevice := 0 to numDevices - 1 do
    begin
      meError := GetDevicePlugged(idxDevice, plugged);

      if (meError = ME_ERRNO_SUCCESS) and (plugged = ME_PLUGGED_IN) then
      begin
        GetSubdevicesByType(idxDevice, ME_TYPE_FI);
      end else
      begin
        memoInfo.Lines.Append('Device ' + IntToStr(idxDevice) + ' not plugged in.');
      end;
    end;
  end;
  DisplaySubdevices();
end;

procedure TfrmMain.DisplaySubdevices();
begin
  if coBoxSubDevs.Items.Count > 0 then
  begin
    memoInfo.Lines.Append('Number of FI sub-devices found: ' + IntToStr(coBoxSubDevs.Items.Count));
    memoInfo.Lines.AddStrings(coBoxSubDevs.Items);
    memoInfo.Lines.Add('');
    coBoxSubDevs.ItemIndex := 0;
    coBoxSubDevsChange(coBoxSubDevs);
    btnDeviceRead.Enabled := True;
    btnReset.Enabled := True;
  end else
  begin
    memoInfo.Lines.Append('No FI sub-devices found in system.');
    CloseMEiDSDriver;
  end;
end;

procedure TfrmMain.GetSubdevicesByType(idxDevice: Integer; subdeviceType: Integer);
var
  idxSubdevice: Integer;
  numSubdevices: Integer;
  meError: Integer;
  idxMatchedSubdevice: Integer;
  devName: array[0..ME_DEVICE_NAME_MAX_COUNT] of AnsiChar;
begin
  meError := meQueryNumberSubdevices(idxDevice, numSubdevices);
  if meError = ME_ERRNO_SUCCESS then
  begin
    meQueryNameDevice(idxDevice, devName, ME_DEVICE_NAME_MAX_COUNT);

    idxSubdevice := 0;
    while idxSubdevice < numSubdevices do
    begin
      meError := meQuerySubdeviceByType(idxDevice,
                                        idxSubdevice,
                                        subdeviceType,
                                        ME_SUBTYPE_ANY,
                                        idxMatchedSubdevice);

      if meError = ME_ERRNO_SUCCESS then
      begin
        AddFiSubdevice(string(devName), idxDevice, idxMatchedSubdevice, subdeviceType);
        idxSubdevice := idxMatchedSubdevice + 1;
      end else
      begin
        Break;
      end;
    end;
  end;
end;

procedure TfrmMain.AddFiSubdevice(devName: string; idxDevice: Integer; idxSubdevice: Integer; subdeviceType: Integer);
var
  numChannels: Integer;
begin
  meQueryNumberChannels(idxDevice,
                        idxSubdevice,
                        numChannels);

  fiSubdevices.AddSubdevice(devName, idxDevice, idxSubdevice, numChannels);

  coBoxSubDevs.Items.Append(devName
                            + ',  Device ' + IntToStr(idxDevice)
                            + ',  Subdevice ' + IntToStr(idxSubdevice)
                            + ',  Number of Channels ' + IntToStr(numChannels)
                            + ',  Type: ' + GetSubDevTypeName(subdeviceType));
end;

procedure TfrmMain.CloseMEiDSDriver();
begin
  if openErrSuccess = ME_ERRNO_SUCCESS then
  begin
    meClose(ME_CLOSE_NO_FLAGS);
    openErrSuccess := ME_IDS_NOTOPENED;
    fiSubdevices.Clear;
    coBoxSubDevs.Clear;
    coBoxChannel.Clear;
    btnDeviceRead.Enabled := False;
    btnReset.Enabled := False;
  end;
end;

procedure TfrmMain.btnDeviceReadClick(Sender: TObject);
var
  meError: Integer;
  flags: Integer;
  idxChannel: Integer;
  timeout: Integer;
  fiSubdevice: TFiSubdevice;
  valueTotal: Integer;
  value1stPhase: Integer;
begin
  try
    timeout := StrToInt(editTimeout.Text);
  except
    on e: Exception do
    begin
      MessageDlg('Error: ' + e.Message, mtError, [mbOK], 0);
      editTimeout.SetFocus;
      Exit;
    end;
  end;

  if (coBoxSubdevs.ItemIndex >= 0) then
  begin
    fiSubdevice := fiSubdevices.Items[coBoxSubDevs.ItemIndex];
    idxChannel := coBoxChannel.ItemIndex;

    meError := ConfigureSubdeviceForInput(fiSubdevice, idxChannel);
    if meError = ME_ERRNO_SUCCESS then
    begin
      flags := ME_IO_SINGLE_TYPE_FIO_TICKS_TOTAL;
      meError := ReadSingleValue(fiSubdevice, idxChannel, timeout, flags, valueTotal);

      if meError = ME_ERRNO_SUCCESS then
      begin
        flags := ME_IO_SINGLE_TYPE_FI_LAST_VALUE or ME_IO_SINGLE_TYPE_FIO_TICKS_FIRST_PHASE;
        meError := ReadSingleValue(fiSubdevice, idxChannel, timeout, flags, value1stPhase);

        if meError = ME_ERRNO_SUCCESS then
        begin
          meError := DisplayValues(fiSubdevice, idxChannel, valueTotal, value1stPhase);
        end;
      end;
      if meError <> ME_ERRNO_SUCCESS then
      begin
        DisplayErrorMessage('Error: ', meError);
      end;
    end;
  end;
end;

function TfrmMain.ConfigureSubdeviceForInput(fiSubdevice: TFiSubdevice; idxChannel: Integer): Integer;
begin
  Result := meIOSingleConfig(fiSubdevice.IndexDevice,
                              fiSubdevice.IndexSubdevice,
                              idxChannel,
                              ME_SINGLE_CONFIG_FIO_INPUT,
                              ME_VALUE_NOT_USED,
                              ME_TRIG_CHAN_DEFAULT,
                              ME_TRIG_TYPE_NONE,
                              ME_VALUE_NOT_USED,
                              ME_IO_SINGLE_CONFIG_NO_FLAGS);
end;

function TfrmMain.ReadSingleValue(fiSubdevice: TFiSubdevice; idxChannel: Integer;
  timeOut: Integer; flags: Integer; var value: Integer): Integer;
var
  meError: Integer;
  io_single: array[0..0] of TMeIOSingle;
begin
  io_single[0].iDevice := fiSubdevice.IndexDevice;
  io_single[0].iSubDevice := fiSubdevice.IndexSubdevice;
  io_single[0].iChannel := idxChannel;
  io_single[0].iDir := ME_DIR_INPUT;
  io_single[0].iValue := 0;
  io_single[0].iTimeOut := timeOut;
  io_single[0].iFlags := flags;

  meError := meIOSingle(io_single[0],
                        1,
                        ME_IO_SINGLE_NO_FLAGS);

  value := io_single[0].iValue;
  Result := meError;
end;

function TfrmMain.CalculateTime(fiSubdevice: TFiSubdevice; timer: Integer;
  valueDigital: Integer; var time: Double): Integer;
begin
  Result := meIOSingleTicksToTime(fiSubdevice.IndexDevice,
                                  fiSubdevice.IndexSubdevice,
                                  timer,
                                  valueDigital,
                                  0,
                                  time,
                                  ME_IO_SINGLE_TICKS_TO_TIME_NO_FLAGS);
end;



function TfrmMain.DisplayValues(fiSubdevice: TFiSubdevice; idxChannel: Integer;
  valueTotal: Integer; value1stPhase: Integer): Integer;
var
  meError: Integer;
  firstPhaseTime: Double;
  periodicTime: Double;
const
  FORMAT_STR = '0.##########';
begin
  meError := CalculateTime(fiSubdevice, ME_TIMER_FIO_TOTAL, valueTotal, periodicTime);
  if meError = ME_ERRNO_SUCCESS then
  begin
    CalculateTime(fiSubdevice, ME_TIMER_FIO_FIRST_PHASE, value1stPhase, firstPhaseTime);
    Inc(lineNumber);
    memoInfo.Lines.Add(Format('%.2d', [lineNumber])
                       + ' Read from ' + fiSubdevice.DeviceName
                       + ',  Device: ' + IntToStr(fiSubdevice.IndexDevice)
                       + ',  Subdevice: ' + IntToStr(fiSubdevice.IndexSubdevice)
                       + ',  Channel: ' + IntToStr(idxChannel));

    memoInfo.Lines.Add('  Ticks total: ' + IntToStr(valueTotal)
                       + ',  seconds: ' + FormatFloat(FORMAT_STR, periodicTime)
                       + ',  Hz: ' + FormatFloat(FORMAT_STR, 1 / periodicTime));
    memoInfo.Lines.Add('  Ticks 1st phase: ' + IntToStr(value1stPhase)
                       + ',  seconds: ' + FormatFloat(FORMAT_STR, firstPhaseTime)
                       + ',  Duty cycle: ' + FormatFloat('0.###', (valueTotal / value1stPhase) * 100)
                       + ' %');
  end;
  Result := meError;
end;

procedure TfrmMain.coBoxSubdevsChange(Sender: TObject);
var
  fiSubDevice: TFiSubdevice;
  i: Integer;
begin
  coBoxChannel.Clear;
  fiSubDevice := fiSubdevices.Items[coBoxSubDevs.itemIndex];
  for i := 0 to fiSubDevice.NumberOfChannels -1 do
  begin
    coBoxChannel.Items.Add(IntToStr(i));
  end;
  if coBoxChannel.Items.Count > 0 then coBoxChannel.ItemIndex := 0;
end;

procedure TfrmMain.btnResetClick(Sender: TObject);
var
  meError: Integer;
  idxDevice: Integer;
  idxSubDevice: Integer;
  fiSubdevice: TFiSubdevice;
begin
  if coBoxSubDevs.ItemIndex >= 0 then
  begin
    fiSubdevice := fiSubdevices.Items[coBoxSubDevs.ItemIndex];
    idxDevice := fiSubdevice.IndexDevice;
    idxSubDevice := fiSubdevice.IndexSubdevice;

    meError := meIOResetSubdevice(idxDevice,
                                  idxSubDevice,
                                  ME_IO_RESET_SUBDEVICE_NO_FLAGS);
    if meError = ME_ERRNO_SUCCESS then
    begin
      memoInfo.Lines.Add('Reset Device ' + IntToStr(idxDevice) + ', Subdevice ' + IntToStr(idxSubDevice));
    end else
    begin
      DisplayErrorMessage('meIOResetSubdevice - Error: ', meError);
    end;
  end;
end;

procedure TfrmMain.DisplayVersionLibrary;
var
  ver: Integer;
begin
  meQueryVersionLibrary(ver);
  memoInfo.Lines.Append('Library version: ' + MeVersionToString(ver));
end;

procedure TfrmMain.DisplayVersionMainDriver;
var
  ver: Integer;
begin
  meQueryVersionMainDriver(ver);
  memoInfo.Lines.Append('Main driver version: ' + MeVersionToString(ver));
end;

function TfrmMain.GetDevicePlugged(idxDevice: Integer; var plugged: Integer): Integer;
var
  busNo: Integer;
  vendorId: Integer;
  deviceId: Integer;
  funcNo: Integer;
  busType: Integer;
  serialNo: Integer;
  devNo: Integer;
begin
  Result := meQueryInfoDevice(idxDevice,
                    vendorId,
                    deviceId,
                    serialNo,
                    busType,
                    busNo,
                    devNo,
                    funcNo,
                    plugged);
end;

function TfrmMain.GetSubDevTypeName(subdeviceType: Integer): string;
begin
  case subdeviceType of
    ME_TYPE_INVALID: Result := 'ME_TYPE_INVALID';
    ME_TYPE_AI: Result := 'ME_TYPE_AI';
    ME_TYPE_AO: Result := 'ME_TYPE_AO';
    ME_TYPE_DIO: Result := 'ME_TYPE_DIO';
    ME_TYPE_DO: Result := 'ME_TYPE_DO';
    ME_TYPE_DI: Result := 'ME_TYPE_DI';
    ME_TYPE_CTR: Result := 'ME_TYPE_CTR';
    ME_TYPE_EXT_IRQ: Result := 'ME_TYPE_EXT_IRQ';
    ME_TYPE_FIO: Result := 'ME_TYPE_FIO';
    ME_TYPE_FO: Result := 'ME_TYPE_FO';
    ME_TYPE_FI: Result := 'ME_TYPE_FI';
    ME_TYPE_FPGA: Result := 'ME_TYPE_FPGA';
  else
    Result := '$' + IntToHex(subdeviceType, 8);
  end;
end;

procedure TfrmMain.DisplayErrorMessage(ErrMsg: string; meError: Integer);
var
  errorMsg: array[0..ME_ERROR_MSG_MAX_COUNT] of AnsiChar;
begin
  meErrorGetMessage(meError, errorMsg, ME_ERROR_MSG_MAX_COUNT);
  memoInfo.lines.add(ErrMsg + IntToStr(meError));
  memoInfo.Lines.Add('ErrorMessage: ' + errorMsg);
end;

function TfrmMain.MeVersionToString(version: Integer): string;
begin
  Result := Format('%d.%d.%d.%d', [version shr 24, $FF and (version shr 16), $FF and (version shr 8), $FF and version]);
end;

procedure TfrmMain.btnClearHwInfoClick(Sender: TObject);
begin
  memoInfo.Clear;
  lineNumber := 0;
end;

procedure TfrmMain.btnCloseClick(Sender: TObject);
begin
  Close;
end;

procedure TfrmMain.btnAboutClick(Sender: TObject);
begin
  frmAbout := TfrmAbout.Create(Self);
  try
    frmAbout.PopupParent := Self;
    frmAbout.ShowModal;
  finally
    frmAbout.Free;
  end;
end;

end.
