﻿'==============================================================================
'Copyright(c), Meilhaus Electronic GmbH
'              http://www.meilhaus.de
'              support@meilhaus.de
'
'Source File : FormMain.vb
'Description : Example program for detecting bit changes (e.g. ME-8100, ME-8200, ME-5100)
'              - Open ME-iDS main driver
'              - Go through the list of all devices and find subdevices with Bit change capability
'              - Add matching subdevices to the combobox, so the user can select one
'              - Close ME-iDS main driver when program is closed
'==============================================================================
'Bitchange:
' This program can only work (detect changes) when an external signal is provided at the pins of the selected subdevice
' e.g. use of a bounce-free switch
' Two options are supported to get informed about a bitchange on a subdevice
' - Use meIOIrqSetCallback and a user defined callback function
'   Using a callback function allows the program to respond to other windows messages while waiting for a bitchange to occur
'   Note: The callback function is called from a different thread created by the ME-iDS library
' - Use meIOIrqWait means wait explicitly for an interrupt to arrive 
'   No extra thread is created by the ME-iDS library.
'   When the wait function (meIOIrqWait) is used and the Timeout is set to indefinitely (iTimeout = 0)
'   this program can't respond until a bitchange is detected, because it doesn't use a seperate thread for the user interface.
'   After the bitchange occured, meIOIrqStop is called. 
'   Click on IRQ Start again to detect another bitchange

Imports System.Text
Imports System.Runtime.InteropServices

Public Class frmMain
    Private openErrSuccess As Integer = Constants.ME_IDS_NOTOPENED
    Private lineNumber As Integer
    Private subdeviceList As New List(Of BitChSubdevice)
    Dim IrqCallback As New meIDS.meIOIrqCB_t(AddressOf meIOIrqCB)
    Private objectHandle As New GCHandle()
    Dim bitChBoxes(0 To 31) As CheckBox
    Dim bitIndicators(0 To 31) As Label
    Private Delegate Sub DisplayValuesDelegate(idxDevice As Integer, idxSubdevice As Integer, idxChannel As Integer, _
                                         irqCount As Integer, value As Integer, context As IntPtr, errorCode As Integer)

    Private Sub frmMain_Shown(sender As System.Object, e As System.EventArgs) Handles MyBase.Shown
        InitBitMaskCheckBoxArray()
        InitBitIndicatorsArray()
        FillStatusFlags()

        OpenMEiDSDriver()
    End Sub

    Private Sub btnClose_Click(sender As System.Object, e As System.EventArgs) Handles btnClose.Click
        Close()
    End Sub

    Private Sub frmMain_FormClosing(sender As System.Object, e As System.Windows.Forms.FormClosingEventArgs) Handles MyBase.FormClosing
        CloseMEiDSDriver()
    End Sub

    Private Sub OpenMEiDSDriver()
        Try
            textBoxInfo.AppendText("Opening Maindriver..." & Environment.NewLine)
            openErrSuccess = meIDS.meOpen(meIDS.ME_OPEN_NO_FLAGS)

            If openErrSuccess = meIDS.ME_ERRNO_SUCCESS Then
                DisplayVersionLibrary()
                DisplayVersionMainDriver()

                textBoxInfo.AppendText(Environment.NewLine)

                Dim numDevices As Integer
                meIDS.meQueryNumberDevices(numDevices)
                textBoxInfo.AppendText("Number of devices found: " & numDevices.ToString() & Environment.NewLine)

                GetSubdevicesWithBitchange(numDevices)
            Else
                DisplayErrorMessage("meOpen - Error: ", openErrSuccess)
            End If
        Catch ex As DllNotFoundException
            textBoxInfo.AppendText("The ME-iDS is not properly installed." & Environment.NewLine)
            textBoxInfo.AppendText(ex.Message & Environment.NewLine)
        End Try
    End Sub

    Private Sub GetSubdevicesWithBitchange(numDevices As Integer)
        Dim idxDevice As Integer
        If openErrSuccess = meIDS.ME_ERRNO_SUCCESS AndAlso numDevices > 0 Then
            textBoxInfo.AppendText("Retrieving sub-devices of type ME_TYPE_DI (Digital In) and ME_TYPE_DIO (Digital In/Out) and bitchange capability..." & Environment.NewLine)

            For idxDevice = 0 To numDevices - 1
                Dim meError As Integer
                Dim plugged As Integer
                meError = GetDevicePlugged(idxDevice, plugged)
                If meError = meIDS.ME_ERRNO_SUCCESS AndAlso plugged = meIDS.ME_PLUGGED_IN Then
                    GetSubdevicesByType(idxDevice, meIDS.ME_TYPE_DI)
                    GetSubdevicesByType(idxDevice, meIDS.ME_TYPE_DIO)
                Else
                    textBoxInfo.AppendText("Device " & idxDevice.ToString() & " not plugged in" & Environment.NewLine)
                End If
            Next
        End If
        DisplaySubdevices()
    End Sub

    Private Sub GetSubdevicesByType(idxDevice As Integer, subdeviceType As Integer)
        Dim meError As Integer
        Dim numSubdevices As Integer
        Dim idxMatchedSubdevice As Integer
        Dim deviceName = New StringBuilder(meIDS.ME_DEVICE_NAME_MAX_COUNT)
        meError = meIDS.meQueryNumberSubdevices(idxDevice, numSubdevices)
        If (meError = meIDS.ME_ERRNO_SUCCESS) Then
            Dim idxSubdevice As Integer = 0
            meIDS.meQueryNameDevice(idxDevice, deviceName, meIDS.ME_DEVICE_NAME_MAX_COUNT)
            While (idxSubdevice < numSubdevices)

                meError = meIDS.meQuerySubdeviceByType(idxDevice,
                                                       idxSubdevice,
                                                       subdeviceType,
                                                       meIDS.ME_SUBTYPE_ANY,
                                                        idxMatchedSubdevice)
                If (meError = meIDS.ME_ERRNO_SUCCESS) Then
                    AddBitchangeSubdevice(deviceName.ToString(), idxDevice, idxMatchedSubdevice, subdeviceType)
                    idxSubdevice = idxMatchedSubdevice + 1
                Else
                    Exit While
                End If
            End While
        End If
    End Sub

    Private Sub AddBitchangeSubdevice(devName As String, idxDevice As Integer, idxSubdevice As Integer, subdeviceType As Integer)
        Dim subdevCaps As Integer
        Dim desiredCaps As Integer = GetBitChangeCaps()
        Dim meError As Integer = meIDS.meQuerySubdeviceCaps(idxDevice,
                                             idxSubdevice,
                                              subdevCaps)
        If ((meError = meIDS.ME_ERRNO_SUCCESS) AndAlso ((subdevCaps And desiredCaps) <> 0)) Then
            Dim numOfChannels As Integer
            meIDS.meQueryNumberChannels(idxDevice, idxSubdevice, numOfChannels)
            Dim bitChSubdev As BitChSubdevice = New BitChSubdevice(devName, idxDevice, idxSubdevice, numOfChannels, subdevCaps)
            subdeviceList.Add(bitChSubdev)

            coBoxSubDevs.Items.Add(devName.ToString() _
                                    & ",  Device " & idxDevice.ToString() _
                                    & ",  Subdevice " & idxSubdevice.ToString() _
                                    & ",  Type: " & GetSubDevTypeName(subdeviceType) _
                                    & ",  Channels: " & numOfChannels.ToString())
        End If
    End Sub

    Private Shared Function GetBitChangeCaps() As Integer
        Dim desiredCaps As Integer = meIDS.ME_CAPS_DIO_BIT_PATTERN_IRQ _
                                    Or meIDS.ME_CAPS_DIO_BIT_MASK_IRQ_EDGE_RISING _
                                    Or meIDS.ME_CAPS_DIO_BIT_MASK_IRQ_EDGE_FALLING _
                                    Or meIDS.ME_CAPS_DIO_BIT_MASK_IRQ_EDGE_ANY
        Return desiredCaps
    End Function
    Private Sub FillSubdeviceCaps(idxSelectedSubdev As Integer)
        If idxSelectedSubdev >= 0 Then
            Dim bitChSubdev As BitChSubdevice = subdeviceList(idxSelectedSubdev)
            coBoxEdges.Items.Clear()
            If (bitChSubdev.subdeviceCaps And meIDS.ME_CAPS_DIO_BIT_MASK_IRQ_EDGE_RISING) = meIDS.ME_CAPS_DIO_BIT_MASK_IRQ_EDGE_RISING Then
                coBoxEdges.Items.Add(New ComboBoxItem("Rising", meIDS.ME_IRQ_EDGE_RISING))
            End If
            If (bitChSubdev.subdeviceCaps And meIDS.ME_CAPS_DIO_BIT_MASK_IRQ_EDGE_FALLING) = meIDS.ME_CAPS_DIO_BIT_MASK_IRQ_EDGE_FALLING Then
                coBoxEdges.Items.Add(New ComboBoxItem("Falling", meIDS.ME_IRQ_EDGE_FALLING))
            End If
            If (bitChSubdev.subdeviceCaps And meIDS.ME_CAPS_DIO_BIT_MASK_IRQ_EDGE_ANY) = meIDS.ME_CAPS_DIO_BIT_MASK_IRQ_EDGE_ANY Then
                coBoxEdges.Items.Add(New ComboBoxItem("Any", meIDS.ME_IRQ_EDGE_ANY))
            End If
            coBoxEdges.SelectedIndex = coBoxEdges.Items.Count - 1

            UncheckBitMask()
            ClearBitIndicators()
        End If
    End Sub

    Private Sub DisplayIrqValues(idxDevice As Integer, idxSubdevice As Integer, idxChannel As Integer, _
                                     irqCount As Integer, value As Integer, context As IntPtr, errorCode As Integer)
        lineNumber = lineNumber + 1
        If (Me.InvokeRequired) Then
            Dim dispDelegate As New DisplayValuesDelegate(AddressOf DisplayIrqValues)
            Me.Invoke(dispDelegate, New Object() {idxDevice, idxSubdevice, idxChannel, irqCount, value, context, errorCode})
        Else
            DisplayValues(idxDevice, idxSubdevice, value)
        End If
    End Sub

    Private Sub DisplayValues(idxDevice As Integer, idxSubdevice As Integer, value As Integer)
        textBoxInfo.AppendText(String.Format("{0:00}", lineNumber) _
                    & " Bitchange on Device: " & idxDevice.ToString() _
                    & ",  Subdevice: " & idxSubdevice.ToString() _
                    & ",  Value hex: " & String.Format("0x{0:X}", value) _
                    & ",  bin: " & IntToBin(value, 32) _
                    & ",  dec: " & value.ToString() _
                    & Environment.NewLine)
        DisplayBitIndicators(value)
    End Sub

    Private Shared Function meIOIrqCB(idxDevice As Integer, idxSubdevice As Integer, idxChannel As Integer, _
                                         irqCount As Integer, value As Integer, pContext As IntPtr, errorCode As Integer) As Integer
        If (errorCode = meIDS.ME_ERRNO_SUCCESS) AndAlso (pContext <> IntPtr.Zero) Then
            Dim objectHandle As GCHandle = GCHandle.FromIntPtr(pContext)
            Dim fm As frmMain = objectHandle.Target
            If Not fm Is Nothing Then
                fm.DisplayIrqValues(idxDevice, idxSubdevice, idxChannel, irqCount, value, pContext, errorCode)
            End If
        End If
        Return errorCode
    End Function

    Private Sub btnResetSubdev_Click(sender As System.Object, e As System.EventArgs) Handles btnResetSubdev.Click
        Dim bitChSubdev As BitChSubdevice
        Dim meError As Integer
        bitChSubdev = subdeviceList(coBoxSubDevs.SelectedIndex)
        meError = meIDS.meIOResetSubdevice(bitChSubdev.deviceIndex, _
                                               bitChSubdev.subdeviceIndex, _
                                               meIDS.ME_IO_RESET_SUBDEVICE_NO_FLAGS)
        If (meError = meIDS.ME_ERRNO_SUCCESS) Then
            textBoxInfo.AppendText("Reset Device " & bitChSubdev.deviceIndex.ToString() & ", Subdevice " & bitChSubdev.subdeviceIndex.ToString() & Environment.NewLine)
            EnableUI(True)
        Else
            DisplayErrorMessage("Error reset Subdevice", meError)
        End If
        ClearBitIndicators()
    End Sub

    Private Sub DisplayErrorMessage(message As String, meError As Integer)
        Dim errMessage = New StringBuilder(meIDS.ME_ERROR_MSG_MAX_COUNT)
        textBoxInfo.AppendText(message & meError.ToString() & Environment.NewLine)
        meIDS.meErrorGetMessage(meError, errMessage, meIDS.ME_ERROR_MSG_MAX_COUNT)
        textBoxInfo.AppendText("ErrorMessage: " & errMessage.ToString & Environment.NewLine)
    End Sub

    Private Sub CloseMEiDSDriver()
        If openErrSuccess = meIDS.ME_ERRNO_SUCCESS Then
            If (coBoxSubDevs.SelectedIndex >= 0) Then
                Dim bitChSubdev As BitChSubdevice = subdeviceList(coBoxSubDevs.SelectedIndex)
                meIDS.meIOIrqStop(bitChSubdev.deviceIndex, bitChSubdev.subdeviceIndex, 0, meIDS.ME_IO_IRQ_STOP_NO_FLAGS)
            End If
            If (objectHandle.IsAllocated) Then
                objectHandle.Free()
            End If
            meIDS.meClose(meIDS.ME_CLOSE_NO_FLAGS)
            openErrSuccess = Constants.ME_IDS_NOTOPENED
            btnResetSubdev.Enabled = False
            EnableUI(False)
            subdeviceList.Clear()
            coBoxSubDevs.Items.Clear()
            coBoxEdges.Items.Clear()
        End If
    End Sub

    Private Sub coBoxSubDevs_SelectedIndexChanged(sender As System.Object, e As System.EventArgs) Handles coBoxSubDevs.SelectedIndexChanged
        FillSubdeviceCaps(coBoxSubDevs.SelectedIndex)
    End Sub

    Private Sub btnStartIRQ_Click(sender As System.Object, e As System.EventArgs) Handles btnStartIRQ.Click
        If coBoxSubDevs.SelectedIndex >= 0 Then

            Dim meError As Integer
            Dim bitChSubdev As BitChSubdevice = subdeviceList(coBoxSubDevs.SelectedIndex)

            EnableUI(False)
            Try
                ConfigureSubdeviceForInput(bitChSubdev)

                If (rbUseCallbackFunc.Checked) Then
                    SetIrqCallbackFunction(bitChSubdev)
                    meError = StartIrq(bitChSubdev)
                    If (meError <> meIDS.ME_ERRNO_SUCCESS) Then
                        DisplayErrorMessage("meIOIrqStart Error:", meError)
                        EnableUI(True)
                    End If
                ElseIf (rbUseIRQWaitFunc.Checked) Then
                    If (objectHandle.IsAllocated) Then
                        objectHandle.Free()
                        RemoveCallbackFunction(bitChSubdev)
                    End If

                    Dim iTimeout As Integer = GetTimeoutValue()
                    meError = StartIrq(bitChSubdev)
                    If (meError = meIDS.ME_ERRNO_SUCCESS) Then
                        bitChSubdev.timeOut = iTimeout
                        BackgroundWorker1.RunWorkerAsync(bitChSubdev)
                    End If
                    If (meError <> meIDS.ME_ERRNO_SUCCESS) Then
                        DisplayErrorMessage("Error:", meError)
                    End If
                End If
            Catch aEx As System.ArgumentException
                MessageBox.Show("Error: " & aEx.Message)
                EnableUI(True)
            Catch ex As System.Exception
                MessageBox.Show("Error: " & ex.Message)
                EnableUI(True)
            End Try
        End If
    End Sub

    Private Function StartIrq(bitChSubdev As BitChSubdevice) As Integer
        Dim meError As Integer
        Dim irqArg As Integer
        Dim irqEdge As Integer
        Dim irqStatusFlag As Integer

        irqArg = GetEdgesMask()
        irqEdge = coBoxEdges.SelectedItem.GetValue
        irqStatusFlag = coBoxStatusFlag.SelectedItem.GetValue

        DisplayEdgesAndBitMask(irqArg)
        meError = meIDS.meIOIrqStart(bitChSubdev.deviceIndex,
                                         bitChSubdev.subdeviceIndex,
                                         0,
                                         meIDS.ME_IRQ_SOURCE_DIO_MASK,
                                         irqEdge,
                                         irqArg,
                                         irqStatusFlag)
        Return meError
    End Function

    Private Sub DisplayEdgesAndBitMask(ByVal irqArg As Integer)
        textBoxInfo.AppendText("StartIRQ set Edges '" & coBoxEdges.SelectedItem.ToString() _
                     & "', BitMask: " & IntToBin(irqArg, 32) & Environment.NewLine)
    End Sub

    Private Sub SetIrqCallbackFunction(bitChSubdev As BitChSubdevice)
        If (Not objectHandle.IsAllocated) Then
            objectHandle = GCHandle.Alloc(Me)
        End If
        meIDS.meIOIrqSetCallback(bitChSubdev.deviceIndex,
                                 bitChSubdev.subdeviceIndex,
                                 IrqCallback,
                                 objectHandle,
                                 meIDS.ME_IO_IRQ_SET_CALLBACK_NO_FLAGS)
    End Sub
    Private Sub BackgroundWorker1_DoWork(sender As System.Object, e As System.ComponentModel.DoWorkEventArgs) Handles BackgroundWorker1.DoWork
        Dim value As Integer
        Dim irqCount As Integer
        Dim meError As Integer
        Dim bitChSubdev As BitChSubdevice = e.Argument
        bitChSubdev.meErrorCode = meIDS.ME_ERRNO_SUCCESS

        While (True)
            meError = meIDS.meIOIrqWait(bitChSubdev.deviceIndex,
                                        bitChSubdev.subdeviceIndex,
                                        0,
                                        irqCount,
                                        value,
                                        bitChSubdev.timeOut,
                                        meIDS.ME_IO_IRQ_WAIT_NO_FLAGS)
            If (meError = meIDS.ME_ERRNO_SUCCESS) Then
                DisplayIrqValues(bitChSubdev.deviceIndex, bitChSubdev.subdeviceIndex, 0, irqCount, value, IntPtr.Zero, 0)
            Else
                bitChSubdev.meErrorCode = meError
                Exit While
            End If
        End While

        meIDS.meIOIrqStop(bitChSubdev.deviceIndex,
                          bitChSubdev.subdeviceIndex,
                          0,
                          meIDS.ME_IO_IRQ_STOP_NO_FLAGS)
        e.Result = bitChSubdev

    End Sub

    Private Sub BackgroundWorker1_RunWorkerCompleted(sender As System.Object, e As System.ComponentModel.RunWorkerCompletedEventArgs) Handles BackgroundWorker1.RunWorkerCompleted
        Dim bitChSubdev As BitChSubdevice = e.Result
        EnableUI(True)
        If (bitChSubdev.meErrorCode <> meIDS.ME_ERRNO_SUCCESS) Then
            DisplayErrorMessage("meIOIrqWait Error: ", bitChSubdev.meErrorCode)
            bitChSubdev.meErrorCode = meIDS.ME_ERRNO_SUCCESS
        End If
    End Sub

    Private Function GetTimeoutValue() As Integer
        Dim iTimeout As Integer = Integer.Parse(numUDTimeout.Text)
        Return iTimeout
    End Function

    Private Sub RemoveCallbackFunction(bitChSubdev As BitChSubdevice)
        meIDS.meIOIrqSetCallback(bitChSubdev.deviceIndex, _
                                bitChSubdev.subdeviceIndex, _
                                Nothing, _
                                IntPtr.Zero, _
                                meIDS.ME_IO_IRQ_SET_CALLBACK_NO_FLAGS)
    End Sub

    Private Function GetEdgesMask() As Integer
        Dim irqArg As Integer = 0
        Dim i As Integer
        For i = 31 To 0 Step -1
            If bitChBoxes(i).Checked Then
                irqArg = irqArg Or (1 << i)
            End If
        Next
        If (irqArg = 0) Then
            Throw New System.ArgumentException("Please check the bits in the Edges Bit Mask that will detect a bit change.")
        End If

        Return irqArg
    End Function

    Private Sub ConfigureSubdeviceForInput(bitChSubdev As BitChSubdevice)
        Dim meError As Integer = meIDS.meIOSingleConfig(bitChSubdev.deviceIndex,
                                             bitChSubdev.subdeviceIndex,
                                             0, _
                                             meIDS.ME_SINGLE_CONFIG_DIO_INPUT,
                                             meIDS.ME_REF_NONE, _
                                             meIDS.ME_TRIG_CHAN_NONE,
                                             meIDS.ME_TRIG_TYPE_NONE,
                                             meIDS.ME_TRIG_EDGE_NONE,
                                             bitChSubdev.singleConfigFlag)
    End Sub

    Private Sub btnStopIRQ_Click(sender As System.Object, e As System.EventArgs) Handles btnStopIRQ.Click
        StopIrq()
    End Sub

    Private Sub StopIrq()
        If (openErrSuccess = meIDS.ME_ERRNO_SUCCESS AndAlso coBoxSubDevs.SelectedIndex >= 0) Then
            Dim bitChSubdev As BitChSubdevice = subdeviceList(coBoxSubDevs.SelectedIndex)
            meIDS.meIOIrqStop(bitChSubdev.deviceIndex, bitChSubdev.subdeviceIndex, 0, meIDS.ME_IO_IRQ_STOP_NO_FLAGS)
            EnableUI(True)
            textBoxInfo.AppendText("Stop IRQ on Device " & bitChSubdev.deviceIndex.ToString() & ", Subdevice " & bitChSubdev.subdeviceIndex.ToString() & Environment.NewLine)
        End If
    End Sub

    Private Sub btnClearIndicator_Click(sender As System.Object, e As System.EventArgs) Handles btnClearIndicator.Click
        ClearBitIndicators()
    End Sub

    Private Sub ClearBitIndicators()
        Dim i As Integer
        For i = 0 To 31
            bitIndicators(i).BackColor = System.Drawing.Color.White
        Next
    End Sub

    Private Sub DisplayBitIndicators(bitValue As Integer)
        Dim bitPos As Integer = 1
        Dim i As Integer
        For i = 0 To 31
            If ((bitValue And bitPos) = bitPos) Then
                bitIndicators(i).BackColor = System.Drawing.Color.Green
            Else
                bitIndicators(i).BackColor = System.Drawing.Color.White
            End If
            bitPos <<= 1
        Next
    End Sub

    Private Sub btnSelectAll_Click(sender As System.Object, e As System.EventArgs) Handles btnSelectAll.Click
        Dim i As Integer
        For i = 0 To 31
            bitChBoxes(i).Checked = True
        Next
    End Sub

    Private Sub btnSelectNone_Click(sender As System.Object, e As System.EventArgs) Handles btnSelectNone.Click
        UncheckBitMask()
    End Sub

    Private Sub UncheckBitMask()
        Dim i As Integer
        For i = 0 To 31
            bitChBoxes(i).Checked = False
        Next
    End Sub

    Private Sub btnSelectInvert_Click(sender As System.Object, e As System.EventArgs) Handles btnSelectInvert.Click
        Dim i As Integer
        For i = 0 To 31
            bitChBoxes(i).Checked = Not bitChBoxes(i).Checked
        Next
    End Sub

    Private Sub btnSelectLowerBits_Click(sender As System.Object, e As System.EventArgs) Handles btnSelectLowerBits.Click
        If (coBoxSubDevs.SelectedIndex >= 0) Then
            Dim bitChSubdev As BitChSubdevice = subdeviceList(coBoxSubDevs.SelectedIndex)
            Dim numBits As Integer = bitChSubdev.numOfChannels
            Dim i As Integer
            If (numBits > 16) Then
                numBits = 16
            End If
            For i = 0 To numBits - 1
                bitChBoxes(i).Checked = True
            Next
        End If
    End Sub

    Private Sub btnSelectUpperBits_Click(sender As System.Object, e As System.EventArgs) Handles btnSelectUpperBits.Click
        If (coBoxSubDevs.SelectedIndex >= 0) Then
            Dim bitChSubdev As BitChSubdevice = subdeviceList(coBoxSubDevs.SelectedIndex)
            Dim numBits As Integer = bitChSubdev.numOfChannels
            Dim i As Integer
            If (numBits > 16) Then
                numBits = 16
            End If
            For i = 16 To 16 + numBits - 1
                bitChBoxes(i).Checked = True
            Next
        End If
    End Sub

    Private Function GetDevicePlugged(idxDevice As Integer, ByRef plugged As Integer) As Integer

        Dim meError As Integer
        Dim vendorId As Integer
        Dim deviceId As Integer
        Dim serialNo As Integer
        Dim busType As Integer
        Dim busNo As Integer
        Dim devNo As Integer
        Dim funcNo As Integer

        meError = meIDS.meQueryInfoDevice(idxDevice, _
                                    vendorId, _
                                    deviceId, _
                                    serialNo, _
                                    busType, _
                                    busNo, _
                                    devNo, _
                                    funcNo, _
                                    plugged)
        Return meError
    End Function

    Private Sub DisplaySubdevices()
        If coBoxSubDevs.Items.Count > 0 Then
            textBoxInfo.AppendText("Number of sub-devices with bitchange found: " & coBoxSubDevs.Items.Count.ToString() & Environment.NewLine)
            Dim i As Integer
            For i = 0 To coBoxSubDevs.Items.Count - 1
                textBoxInfo.AppendText(coBoxSubDevs.Items(i) & Environment.NewLine)
            Next
            textBoxInfo.AppendText(Environment.NewLine)
            coBoxSubDevs.SelectedIndex = 0
            FillSubdeviceCaps(coBoxSubDevs.SelectedIndex)
            btnResetSubdev.Enabled = True
            EnableUI(True)
        Else
            textBoxInfo.AppendText("No sub-devices with bitchange capability found." & Environment.NewLine)
            CloseMEiDSDriver()
        End If
    End Sub

    Private Sub EnableUI(isEnabled As Boolean)
        'enable or disable user interface
        btnStartIRQ.Enabled = isEnabled
        btnStopIRQ.Enabled = Not isEnabled
        coBoxSubDevs.Enabled = isEnabled
        grpBoxBitMask.Enabled = isEnabled
        btnSelectAll.Enabled = isEnabled
        btnSelectNone.Enabled = isEnabled
        btnSelectInvert.Enabled = isEnabled
        btnSelectUpperBits.Enabled = isEnabled
        btnSelectLowerBits.Enabled = isEnabled
        coBoxEdges.Enabled = isEnabled
        coBoxStatusFlag.Enabled = isEnabled
    End Sub

    Private Sub InitBitIndicatorsArray()
        bitIndicators(0) = lblBit0
        bitIndicators(1) = lblBit1
        bitIndicators(2) = lblBit2
        bitIndicators(3) = lblBit3
        bitIndicators(4) = lblBit4
        bitIndicators(5) = lblBit5
        bitIndicators(6) = lblBit6
        bitIndicators(7) = lblBit7
        bitIndicators(8) = lblBit8
        bitIndicators(9) = lblBit9
        bitIndicators(10) = lblBit10
        bitIndicators(11) = lblBit11
        bitIndicators(12) = lblBit12
        bitIndicators(13) = lblBit13
        bitIndicators(14) = lblBit14
        bitIndicators(15) = lblBit15
        bitIndicators(16) = lblBit16
        bitIndicators(17) = lblBit17
        bitIndicators(18) = lblBit18
        bitIndicators(19) = lblBit19
        bitIndicators(20) = lblBit20
        bitIndicators(21) = lblBit21
        bitIndicators(22) = lblBit22
        bitIndicators(23) = lblBit23
        bitIndicators(24) = lblBit24
        bitIndicators(25) = lblBit25
        bitIndicators(26) = lblBit26
        bitIndicators(27) = lblBit27
        bitIndicators(28) = lblBit28
        bitIndicators(29) = lblBit29
        bitIndicators(30) = lblBit30
        bitIndicators(31) = lblBit31
    End Sub

    Private Sub InitBitMaskCheckBoxArray()
        bitChBoxes(0) = chBoxBit0
        bitChBoxes(1) = chBoxBit1
        bitChBoxes(2) = chBoxBit2
        bitChBoxes(3) = chBoxBit3
        bitChBoxes(4) = chBoxBit4
        bitChBoxes(5) = chBoxBit5
        bitChBoxes(6) = chBoxBit6
        bitChBoxes(7) = chBoxBit7
        bitChBoxes(8) = chBoxBit8
        bitChBoxes(9) = chBoxBit9
        bitChBoxes(10) = chBoxBit10
        bitChBoxes(11) = chBoxBit11
        bitChBoxes(12) = chBoxBit12
        bitChBoxes(13) = chBoxBit13
        bitChBoxes(14) = chBoxBit14
        bitChBoxes(15) = chBoxBit15
        bitChBoxes(16) = chBoxBit16
        bitChBoxes(17) = chBoxBit17
        bitChBoxes(18) = chBoxBit18
        bitChBoxes(19) = chBoxBit19
        bitChBoxes(20) = chBoxBit20
        bitChBoxes(21) = chBoxBit21
        bitChBoxes(22) = chBoxBit22
        bitChBoxes(23) = chBoxBit23
        bitChBoxes(24) = chBoxBit24
        bitChBoxes(25) = chBoxBit25
        bitChBoxes(26) = chBoxBit26
        bitChBoxes(27) = chBoxBit27
        bitChBoxes(28) = chBoxBit28
        bitChBoxes(29) = chBoxBit29
        bitChBoxes(30) = chBoxBit30
        bitChBoxes(31) = chBoxBit31
    End Sub

    Private Sub FillStatusFlags()
        coBoxStatusFlag.Items.Add(New ComboBoxItem("Standard", meIDS.ME_IO_IRQ_START_NO_FLAGS))
        coBoxStatusFlag.Items.Add(New ComboBoxItem("Extended", meIDS.ME_IO_IRQ_START_EXTENDED_STATUS))
        coBoxStatusFlag.SelectedIndex = 0
    End Sub

    Private Function GetSubDevTypeName(subdeviceType As Integer)
        Dim tmpType As String
        Select Case subdeviceType
            Case meIDS.ME_TYPE_INVALID
                tmpType = "ME_TYPE_INVALID"
            Case meIDS.ME_TYPE_AI
                tmpType = "ME_TYPE_AI"
            Case meIDS.ME_TYPE_AO
                tmpType = "ME_TYPE_AO"
            Case meIDS.ME_TYPE_DIO
                tmpType = "ME_TYPE_DIO"
            Case meIDS.ME_TYPE_DO
                tmpType = "ME_TYPE_DO"
            Case meIDS.ME_TYPE_DI
                tmpType = "ME_TYPE_DI"
            Case meIDS.ME_TYPE_CTR
                tmpType = "ME_TYPE_CTR"
            Case meIDS.ME_TYPE_EXT_IRQ
                tmpType = "ME_TYPE_EXT_IRQ"
            Case meIDS.ME_TYPE_FIO
                tmpType = "ME_TYPE_FIO"
            Case meIDS.ME_TYPE_FO
                tmpType = "ME_TYPE_FO"
            Case meIDS.ME_TYPE_FI
                tmpType = "ME_TYPE_FI"
            Case meIDS.ME_TYPE_FPGA
                tmpType = "ME_TYPE_FPGA"
            Case Else
                tmpType = String.Format("&H{0:X}", subdeviceType)
        End Select
        Return tmpType
    End Function

    Private Function IntToBin(val As Integer, dispDigits As Integer) As String
        Dim b As Integer
        Dim found As Boolean = False
        Dim tmpStr As String = ""
        For b = 0 To Constants.DIGITNUMS.Length - 1
            If dispDigits = Constants.DIGITNUMS(b) Then
                found = True
                Exit For
            End If
        Next
        If found Then
            If dispDigits = 1 Then
                tmpStr = (val And 1).ToString()
            Else
                For i As Integer = (dispDigits >> 2) - 1 To 0 Step -1
                    b = (val >> (i << 2)) And &HF
                    tmpStr = tmpStr & Constants.BIN_STR(b)
                Next
            End If
        End If
        Return tmpStr
    End Function

    Private Sub DisplayVersionLibrary()
        Dim tmpVersion As Integer
        meIDS.meQueryVersionLibrary(tmpVersion)
        textBoxInfo.AppendText("Library version: " & MeVersionToString(tmpVersion) & Environment.NewLine)
    End Sub

    Private Sub DisplayVersionMainDriver()
        Dim tmpVersion As Integer
        meIDS.meQueryVersionMainDriver(tmpVersion)
        textBoxInfo.AppendText("Maindriver version: " & MeVersionToString(tmpVersion) & Environment.NewLine)
        textBoxInfo.AppendText(Environment.NewLine)
    End Sub

    Private Function MeVersionToString(version As Integer) As String
        Return String.Format("{0}.{1}.{2}.{3}", version >> 24, &HFF And (version >> 16), &HFF And (version >> 8), &HFF And version)
    End Function

    Private Sub btnClearHwInfo_Click(sender As System.Object, e As System.EventArgs) Handles btnClearHwInfo.Click
        textBoxInfo.Clear()
        lineNumber = 0
    End Sub

    Private Sub btnAbout_Click(sender As System.Object, e As System.EventArgs) Handles btnAbout.Click
        AboutBox1.ShowDialog()
    End Sub
End Class

Public Class BitChSubdevice
    Public deviceName As String
    Public deviceIndex As Integer
    Public subdeviceIndex As Integer
    Public numOfChannels As Integer
    Public subdeviceCaps As Integer
    Public singleConfigFlag As Integer
    Public timeOut As Integer
    Public meErrorCode As Integer

    Sub New(DevName As String, DeviceIdx As Integer, SubdevIdx As Integer, NumChannels As Integer, Capabilities As Integer)
        deviceName = DevName
        deviceIndex = DeviceIdx
        subdeviceIndex = SubdevIdx
        numOfChannels = NumChannels
        subdeviceCaps = Capabilities
        SetSingleConfigFlag(NumChannels)
    End Sub

    Private Sub SetSingleConfigFlag(NumberOfChannels As Integer)
        Select Case NumberOfChannels
            Case 1
                singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_BIT
            Case 8
                singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_BYTE
            Case 16
                singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_WORD
            Case 24
                singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_DWORD
            Case 32
                singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_DWORD
            Case Else
                singleConfigFlag = meIDS.ME_VALUE_INVALID
        End Select
    End Sub

End Class

Public Class ComboBoxItem
    Dim dispValue As String
    Dim val As Integer

    Sub New(displayValue As String, value As Integer)
        dispValue = displayValue
        val = value
    End Sub

    Public Overrides Function ToString() As String
        Return dispValue
    End Function

    ReadOnly Property GetValue() As Integer
        Get
            Return val
        End Get
    End Property
End Class

Class Constants
    Public Shared ReadOnly ME_IDS_NOTOPENED As Integer = -1
    Public Shared ReadOnly BIN_STR As String() = {"0000", "0001", "0010", "0011",
                                               "0100", "0101", "0110", "0111",
                                               "1000", "1001", "1010", "1011",
                                               "1100", "1101", "1110", "1111"}
    Public Shared ReadOnly DIGITNUMS As Integer() = {1, 8, 16, 24, 32}
End Class
