﻿'==============================================================================
'Copyright(c), Meilhaus Electronic GmbH
'              http://www.meilhaus.de
'              support@meilhaus.de
'
'Source File : FormMain.vb
'Description : Example program for writing single values to DO and DIO devices
'              - Open ME-iDS main driver
'              - list all currently pluged in ME-iDS devices found in system
'              - determine subdevices of DO and DIO type
'              - Add DO and DIO subdevices to the combobox, so the user can select one
'              - Write data to selected subdevice when user clicks on button "Write Data"
'              - Data are always written to channel 0 with full channel width in this example
'              - Close ME-iDS main driver, if not already done, when program is closed
'==============================================================================
Imports System.Text

Public Class frmMain
    Private openErrSuccess As Integer = Constants.ME_IDS_NOTOPENED
    Private lineNumber As Integer
    Private subdeviceList As New List(Of DoSubdevice)

    Private Sub frmMain_Shown(sender As System.Object, e As System.EventArgs) Handles MyBase.Shown
        FillPortConfiguration()
        OpenMEiDSDriver()
    End Sub

    Private Sub btnClose_Click(sender As System.Object, e As System.EventArgs) Handles btnClose.Click
        Close()
    End Sub

    Private Sub frmMain_FormClosing(sender As System.Object, e As System.Windows.Forms.FormClosingEventArgs) Handles MyBase.FormClosing
        CloseMEiDSDriver()
    End Sub

    Private Sub OpenMEiDSDriver()
        Try
            textBoxInfo.AppendText("Opening Maindriver..." & Environment.NewLine)
            openErrSuccess = meIDS.meOpen(meIDS.ME_OPEN_NO_FLAGS)

            If openErrSuccess = meIDS.ME_ERRNO_SUCCESS Then
                DisplayVersionLibrary()
                DisplayVersionMainDriver()
                textBoxInfo.AppendText(Environment.NewLine)

                Dim numDevices As Integer
                meIDS.meQueryNumberDevices(numDevices)
                textBoxInfo.AppendText("Number of devices found: " & numDevices.ToString() & Environment.NewLine)

                GetDoDioSubdevices(numDevices)
            Else
                DisplayErrorMessage("meOpen - Error: ", openErrSuccess)
            End If
        Catch ex As DllNotFoundException
            textBoxInfo.AppendText("The ME-iDS is not properly installed." & Environment.NewLine)
            textBoxInfo.AppendText(ex.Message & Environment.NewLine)
        End Try
    End Sub

    Private Sub GetDoDioSubdevices(numDevices As Integer)
        Dim idxDevice As Integer

        If openErrSuccess = meIDS.ME_ERRNO_SUCCESS AndAlso numDevices > 0 Then
            textBoxInfo.AppendText("Retrieving sub-devices of type ME_TYPE_DO (Digital Out) and ME_TYPE_DIO (Digital In/Out)..." & Environment.NewLine)
            subdeviceList.Clear()
            coBoxSubDevs.Items.Clear()
            For idxDevice = 0 To numDevices - 1
                Dim meError As Integer
                Dim plugged As Integer

                meError = GetDevicePlugged(idxDevice, plugged)
                If meError = meIDS.ME_ERRNO_SUCCESS AndAlso plugged = meIDS.ME_PLUGGED_IN Then
                    GetSubdevicesByType(idxDevice, meIDS.ME_TYPE_DO)
                    GetSubdevicesByType(idxDevice, meIDS.ME_TYPE_DIO)
                Else
                    textBoxInfo.AppendText(String.Format("Device {0} not plugged in.", idxDevice) & Environment.NewLine)
                End If
            Next
        End If
        DisplaySubdevices()
    End Sub

    Private Sub GetSubdevicesByType(idxDevice As Integer, subdeviceType As Integer)
        Dim meError As Integer
        Dim numSubdevices As Integer
        Dim idxMatchedSubdevice As Integer

        meError = meIDS.meQueryNumberSubdevices(idxDevice, numSubdevices)
        If meError = meIDS.ME_ERRNO_SUCCESS Then
            Dim idxSubdevice As Integer = 0
            Dim deviceName = New StringBuilder(meIDS.ME_DEVICE_NAME_MAX_COUNT)
            meIDS.meQueryNameDevice(idxDevice, deviceName, meIDS.ME_DEVICE_NAME_MAX_COUNT)
            While idxSubdevice < numSubdevices

                meError = meIDS.meQuerySubdeviceByType(idxDevice, _
                                                 idxSubdevice, _
                                                 subdeviceType, _
                                                 meIDS.ME_SUBTYPE_ANY, _
                                                 idxMatchedSubdevice)

                If meError = meIDS.ME_ERRNO_SUCCESS Then
                    AddDoSubdevice(deviceName.ToString(), idxDevice, idxMatchedSubdevice, subdeviceType)
                    idxSubdevice = idxMatchedSubdevice + 1
                Else
                    Exit While
                End If
            End While
        End If
    End Sub

    Private Sub AddDoSubdevice(deviceName As String, idxDevice As Integer, idxSubdevice As Integer, subdeviceType As Integer)
        Dim numOfChannels As Integer
        Dim devCaps As Integer

        meIDS.meQueryNumberChannels(idxDevice, idxSubdevice, numOfChannels)
        meIDS.meQuerySubdeviceCaps(idxDevice, idxSubdevice, devCaps)
        subdeviceList.Add(New DoSubdevice(deviceName.ToString, idxDevice, idxSubdevice, numOfChannels, devCaps))

        coBoxSubDevs.Items.Add(deviceName.ToString _
                               & ",  Device " & idxDevice.ToString() _
                               & ",  Subdevice " & idxSubdevice.ToString() _
                               & ",  Type: " & GetSubDevTypeName(subdeviceType) _
                               & ",  Channels: " & numOfChannels.ToString())
    End Sub

    Private Sub coBoxSubDevs_SelectedIndexChanged(sender As System.Object, e As System.EventArgs) Handles coBoxSubDevs.SelectedIndexChanged
        If (coBoxSubDevs.SelectedIndex >= 0) Then
            Dim doSubdev As DoSubdevice
            doSubdev = subdeviceList(coBoxSubDevs.SelectedIndex)
            coBoxPortConfiguration.Enabled = ((doSubdev.deviceCaps And meIDS.ME_CAPS_DIO_SINK_SOURCE) = meIDS.ME_CAPS_DIO_SINK_SOURCE)
        Else
            coBoxPortConfiguration.Enabled = False
        End If
    End Sub

    Private Sub btnSingleWrite_Click(sender As System.Object, e As System.EventArgs) Handles btnSingleWrite.Click
        Dim doSubdev As DoSubdevice
        doSubdev = subdeviceList(coBoxSubDevs.SelectedIndex)

        Dim valueToWrite As Integer
        Try
            valueToWrite = CheckValueToWrite(txtBoxHexVal.Text, doSubdev.numOfChannels)
            WriteValue(doSubdev, valueToWrite)
        Catch ex As Exception
            MessageBox.Show("Error: " & ex.Message)
            txtBoxHexVal.Focus()
        End Try
    End Sub

    Private Sub btnWriteZero_Click(sender As System.Object, e As System.EventArgs) Handles btnWriteZero.Click
        Dim doSubdev As DoSubdevice
        doSubdev = subdeviceList(coBoxSubDevs.SelectedIndex)
        WriteValue(doSubdev, 0)
    End Sub

    Private Function CheckValueToWrite(valStr As String, numChannels As Integer) As Integer
        Dim x As UInt32
        x = UInt32.Parse(valStr, System.Globalization.NumberStyles.AllowHexSpecifier)
        If (numChannels < 32) Then
            Dim m As Integer
            m = (1 << numChannels) - 1
            If (x > m) Then
                Throw New Exception(String.Format("The value is out of range. Maximum value Hex: {0:X}", m))
            End If
        End If
        Return x
    End Function

    Private Sub WriteValue(doSubdev As DoSubdevice, valueToWrite As Integer)
        Dim meError As Integer

        meError = ConfigureSubdeviceForOutput(doSubdev)
        If (meError = meIDS.ME_ERRNO_SUCCESS) Then
            meError = WriteSingleValue(doSubdev, valueToWrite)
            If (meError = meIDS.ME_ERRNO_SUCCESS) Then
                DisplayValue(doSubdev, valueToWrite)

            Else
                DisplayErrorMessage("meIOSingle - Error: ", meError)
            End If
        Else
            DisplayErrorMessage("meIOSingleConfig - Error: ", meError)
        End If
    End Sub

    Private Function ConfigureSubdeviceForOutput(doSubdev As DoSubdevice) As Integer
        Dim meError As Integer
        If ((doSubdev.deviceCaps And meIDS.ME_CAPS_DIO_SINK_SOURCE) = meIDS.ME_CAPS_DIO_SINK_SOURCE) Then
            Dim singleConfig As Integer = (coBoxPortConfiguration.SelectedItem).GetValue
            meError = meIDS.meIOSingleConfig(doSubdev.deviceIndex,
                                             doSubdev.subdevIndex,
                                             0,
                                             singleConfig,
                                             meIDS.ME_REF_NONE,
                                             meIDS.ME_VALUE_NOT_USED,
                                             meIDS.ME_VALUE_NOT_USED,
                                             meIDS.ME_VALUE_NOT_USED,
                                             doSubdev.singleConfigFlag)

        Else
            meError = meIDS.meIOSingleConfig(doSubdev.deviceIndex,
                                         doSubdev.subdevIndex,
                                         0,
                                         meIDS.ME_SINGLE_CONFIG_DIO_OUTPUT,
                                         meIDS.ME_REF_NONE,
                                         meIDS.ME_TRIG_CHAN_DEFAULT,
                                         meIDS.ME_TRIG_TYPE_NONE,
                                         meIDS.ME_VALUE_NOT_USED,
                                         doSubdev.singleConfigFlag)
        End If
        Return meError
    End Function

    Private Function WriteSingleValue(doSubdev As DoSubdevice, valueToWrite As Integer) As Integer
        Dim io_single(1) As meIDS.meIOSingle_t

        io_single(0).iDevice = doSubdev.deviceIndex
        io_single(0).iSubdevice = doSubdev.subdevIndex
        io_single(0).iChannel = 0
        io_single(0).iDir = meIDS.ME_DIR_OUTPUT
        io_single(0).iValue = valueToWrite
        io_single(0).iTimeOut = 0
        io_single(0).iFlags = doSubdev.singleTypeFlag
        io_single(0).iErrno = 0

        Dim meError As Integer = meIDS.meIOSingle(io_single(0),
                                   1,
                                   meIDS.ME_IO_SINGLE_NO_FLAGS)

        Return meError
    End Function

    Private Sub DisplayValue(doSubdev As DoSubdevice, valueToWrite As Integer)
        lineNumber = lineNumber + 1
        textBoxInfo.AppendText(String.Format("{0:00}", lineNumber) _
                  & " Write to " & doSubdev.deviceName _
                  & ",  Device: " & doSubdev.deviceIndex.ToString() _
                  & ",  Subdevice: " & doSubdev.subdevIndex.ToString() _
                  & ",  Value hex: " & String.Format("0x{0:X}", valueToWrite) _
                  & ",  bin: " & IntToBin(valueToWrite, doSubdev.numOfChannels) _
                  & ",  dec: " & valueToWrite.ToString() + Environment.NewLine)
    End Sub

    Private Sub btnResetSubdev_Click(sender As System.Object, e As System.EventArgs) Handles btnResetSubdev.Click
        Dim doSubdev As DoSubdevice
        Dim meError As Integer

        doSubdev = subdeviceList(coBoxSubDevs.SelectedIndex)
        meError = meIDS.meIOResetSubdevice(doSubdev.deviceIndex, _
                                     doSubdev.subdevIndex, _
                                     meIDS.ME_IO_RESET_SUBDEVICE_NO_FLAGS)
        If meError = meIDS.ME_ERRNO_SUCCESS Then
            textBoxInfo.AppendText("Reset Device " & doSubdev.deviceIndex.ToString() & ", Subdevice " & doSubdev.subdevIndex.ToString() & Environment.NewLine)
        Else
            DisplayErrorMessage("Error resetting Subdevice ", meError)
        End If
    End Sub

    Private Sub FillPortConfiguration()
        coBoxPortConfiguration.Items.Add(New ComboBoxItem("High impedance", meIDS.ME_SINGLE_CONFIG_DIO_HIGH_IMPEDANCE))
        coBoxPortConfiguration.Items.Add(New ComboBoxItem("Sink", meIDS.ME_SINGLE_CONFIG_DIO_SINK))
        coBoxPortConfiguration.Items.Add(New ComboBoxItem("Source", meIDS.ME_SINGLE_CONFIG_DIO_SOURCE))
        coBoxPortConfiguration.SelectedIndex = 0
    End Sub

    Private Function GetDevicePlugged(idxDevice As Integer, ByRef plugged As Integer) As Integer

        Dim meError As Integer
        Dim vendorId As Integer
        Dim deviceId As Integer
        Dim serialNo As Integer
        Dim busType As Integer
        Dim busNo As Integer
        Dim devNo As Integer
        Dim funcNo As Integer

        meError = meIDS.meQueryInfoDevice(idxDevice, _
                                    vendorId, _
                                    deviceId, _
                                    serialNo, _
                                    busType, _
                                    busNo, _
                                    devNo, _
                                    funcNo, _
                                    plugged)
        Return meError
    End Function

    Private Sub DisplaySubdevices()
        If coBoxSubDevs.Items.Count > 0 Then
            textBoxInfo.AppendText("Number of DO or DIO sub-devices found: " & coBoxSubDevs.Items.Count.ToString() & Environment.NewLine)
            Dim i As Integer
            For i = 0 To coBoxSubDevs.Items.Count - 1
                textBoxInfo.AppendText(coBoxSubDevs.Items(i) & Environment.NewLine)
            Next
            textBoxInfo.AppendText(Environment.NewLine)
            coBoxSubDevs.SelectedIndex = 0
            btnSingleWrite.Enabled = True
            btnWriteZero.Enabled = True
            btnResetSubdev.Enabled = True
        Else
            textBoxInfo.AppendText("No DO or DIO sub-devices found." & Environment.NewLine)
            CloseMEiDSDriver()
        End If
    End Sub

    Private Function GetSubDevTypeName(subdeviceType As Integer)
        Dim tmpType As String
        Select Case subdeviceType
            Case meIDS.ME_TYPE_INVALID
                tmpType = "ME_TYPE_INVALID"
            Case meIDS.ME_TYPE_AI
                tmpType = "ME_TYPE_AI"
            Case meIDS.ME_TYPE_AO
                tmpType = "ME_TYPE_AO"
            Case meIDS.ME_TYPE_DIO
                tmpType = "ME_TYPE_DIO"
            Case meIDS.ME_TYPE_DO
                tmpType = "ME_TYPE_DO"
            Case meIDS.ME_TYPE_DI
                tmpType = "ME_TYPE_DI"
            Case meIDS.ME_TYPE_CTR
                tmpType = "ME_TYPE_CTR"
            Case meIDS.ME_TYPE_EXT_IRQ
                tmpType = "ME_TYPE_EXT_IRQ"
            Case meIDS.ME_TYPE_FIO
                tmpType = "ME_TYPE_FIO"
            Case meIDS.ME_TYPE_FO
                tmpType = "ME_TYPE_FO"
            Case meIDS.ME_TYPE_FI
                tmpType = "ME_TYPE_FI"
            Case meIDS.ME_TYPE_FPGA
                tmpType = "ME_TYPE_FPGA"
            Case Else
                tmpType = String.Format("&H{0:X}", subdeviceType)
        End Select
        Return tmpType
    End Function

    Private Sub CloseMEiDSDriver()
        If openErrSuccess = meIDS.ME_ERRNO_SUCCESS Then
            meIDS.meClose(meIDS.ME_CLOSE_NO_FLAGS)
            openErrSuccess = Constants.ME_IDS_NOTOPENED
            coBoxPortConfiguration.Enabled = False
            btnSingleWrite.Enabled = False
            btnWriteZero.Enabled = False
            btnResetSubdev.Enabled = False
            subdeviceList.Clear()
            coBoxSubDevs.Items.Clear()
        End If
    End Sub

    Private Sub DisplayErrorMessage(message As String, meError As Integer)
        Dim errMessage = New StringBuilder(meIDS.ME_ERROR_MSG_MAX_COUNT)
        textBoxInfo.AppendText(message + meError.ToString() & Environment.NewLine)
        meIDS.meErrorGetMessage(meError, errMessage, meIDS.ME_ERROR_MSG_MAX_COUNT)
        textBoxInfo.AppendText("ErrorMessage: " & errMessage.ToString & Environment.NewLine)
    End Sub

    Private Function IntToBin(val As Integer, dispDigits As Integer) As String
        Dim b As Integer
        Dim found As Boolean = False
        Dim tmpStr As String = ""
        For b = 0 To Constants.DIGITNUMS.Length - 1
            If dispDigits = Constants.DIGITNUMS(b) Then
                found = True
                Exit For
            End If
        Next
        If found Then
            If dispDigits = 1 Then
                tmpStr = (val And 1).ToString()
            Else
                For i As Integer = (dispDigits >> 2) - 1 To 0 Step -1
                    b = (val >> (i << 2)) And &HF
                    tmpStr = tmpStr & Constants.BIN_STR(b)
                Next
            End If
        End If
        Return tmpStr
    End Function

    Private Sub DisplayVersionLibrary()
        Dim tmpVersion As Integer
        meIDS.meQueryVersionLibrary(tmpVersion)
        textBoxInfo.AppendText("Library version: " & MeVersionToString(tmpVersion) & Environment.NewLine)
    End Sub

    Private Sub DisplayVersionMainDriver()
        Dim tmpVersion As Integer
        meIDS.meQueryVersionMainDriver(tmpVersion)
        textBoxInfo.AppendText("Maindriver version: " & MeVersionToString(tmpVersion) & Environment.NewLine)
        textBoxInfo.AppendText(Environment.NewLine)
    End Sub

    Private Function MeVersionToString(version As Integer) As String
        Return String.Format("{0}.{1}.{2}.{3}", version >> 24, &HFF And (version >> 16), &HFF And (version >> 8), &HFF And version)
    End Function

    Private Sub btnClearHwInfo_Click(sender As System.Object, e As System.EventArgs) Handles btnClearHwInfo.Click
        textBoxInfo.Clear()
        lineNumber = 0
    End Sub

    Private Sub btnAbout_Click(sender As System.Object, e As System.EventArgs) Handles btnAbout.Click
        AboutBox1.ShowDialog()
    End Sub
End Class

Class DoSubdevice
    Public deviceName As String
    Public deviceIndex As Integer
    Public subdevIndex As Integer
    Public numOfChannels As Integer
    Public deviceCaps As Integer
    Public singleConfigFlag As Integer
    Public singleTypeFlag As Integer

    Sub New(DevName As String, DeviceIdx As Integer, SubdevIdx As Integer, NumChannels As Integer, DevCaps As Integer)
        deviceName = DevName
        deviceIndex = DeviceIdx
        subdevIndex = SubdevIdx
        numOfChannels = NumChannels
        deviceCaps = DevCaps
        SetSingleFlags(NumChannels)
    End Sub

    Private Sub SetSingleFlags(NumberOfChannels As Integer)
        Select Case NumberOfChannels
            Case 1
                singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_BIT
                singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_BIT
            Case 8
                singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_BYTE
                singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_BYTE
            Case 16
                singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_WORD
                singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_WORD
            Case 24
                singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_DWORD
                singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_DWORD
            Case 32
                singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_DWORD
                singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_DWORD
            Case Else
                singleConfigFlag = meIDS.ME_VALUE_INVALID
                singleTypeFlag = meIDS.ME_VALUE_INVALID
        End Select
    End Sub
End Class

Public Class ComboBoxItem
    Dim dispValue As String
    Dim val As Integer

    Sub New(displayValue As String, value As Integer)
        dispValue = displayValue
        val = value
    End Sub

    Public Overrides Function ToString() As String
        Return dispValue
    End Function

    ReadOnly Property GetValue() As Integer
        Get
            Return val
        End Get
    End Property
End Class

Class Constants
    Public Shared ReadOnly ME_IDS_NOTOPENED As Integer = -1

    Public Shared ReadOnly BIN_STR As String() = {"0000", "0001", "0010", "0011",
                                               "0100", "0101", "0110", "0111",
                                               "1000", "1001", "1010", "1011",
                                               "1100", "1101", "1110", "1111"}
    Public Shared ReadOnly DIGITNUMS As Integer() = {1, 8, 16, 24, 32}
End Class

