﻿'==============================================================================
'Copyright(c), Meilhaus Electronic GmbH
'              http://www.meilhaus.de
'              support@meilhaus.de
'
'Source File : FormMain.vb
'Description : Example program for writing to Frequency-out (FO) subdevices
'              - Open main driver (ME-iDS)
'              - Go through the list of all devices and find FO subdevices
'              - Add FO devices to the combobox, so the user can select one
'              - First set the configuration of the selected subdevice
'              - Write data to selected subdevice when user clicks on button "Write Data"
'                The digital signal is emitted
'              - Written values can be read back by click on "Read Data"
'              - Close main driver (ME-iDS) when program is closed
'              - Port configuration for Sink, Souce, High impedance (e.g ME-5810)
'==============================================================================
'  To emit a digital signal two values of the FO subdevice must be set:
'  - 1. "Ticks total" = value of the entire period of the signal
'  - 2. "Ticks first phase" = value of the first phase of the signal (first phase can be high or low)
'    Therefore the function meIOSingle() must be called twice using different flags
'
'  digital signal:
'      +-------------------+         +---
'      |                   |         |   
'      |                   |         |   
'  ----+                   +---------+   
'                                   
'      ^-------Ticks total-----------^
'                                
'      ^--Ticks 1st phase--^
'
'
'  How to start several (a group of) FO-subdevices (FO-channels) synchronously:
'  - configure each subdevice with the flag 'ME_TRIG_CHAN_SYNCHRONOUS'
'  - write the values for 'Ticks total' and 'Ticks first phase' to each channel without the flag 'Synchronous',
'  - set the flag 'ME_TRIG_CHAN_SYNCHRONOUS' for the last subdevice in the group.
'  Port configuration:
'  - Configuration of the output driver of devices with sink, source, high impedance capabilities (e.g. ME-5810A, ME-5002)
'  Configuration:
'    Immediate output = start the output of the configured subdevice(s) or channel(s)
'    after values "Ticks total" and "Ticks first phase" are set
'    Synchronous output = don't start the selected subdevice or channel, set only values for "Ticks total" and "Ticks first phase"
'  - (Also see ME-iDS Help section: Supported Hardware / ME-46xx Description / Special features / Analog output subdevices)
'  Read:
'  - read back the values of the previously emitted signal
'    Ticks total: the value of the entire period
'    Ticks first phase: the value of the first phase of the period
'  Write:
'  - Signal start level: (Start signal with high or low phase)
'    High: first phase of the signal will be high
'    Low: first phase of the signal will he low
'  - Signal start type: (start immediately when button clicked or at the end of the currently emitted signal period
'    Hard: start the output of the signal immediately (on button click)
'    Soft: Controlled start of the new signal when the period of the current signal (if any) is over
'  - Write mode:
'    Direct: This subdevice or channel will be started alone
'    Synchronous: All those subdevices which have been configured with the 'iTrigChan' = ME_TRIG_CHAN_SYNCHRONOUS
'    in meIOSingleConfig above will be started simultaneously together with this subdevice
'------------------------------------------------------------------------------
Imports System.Text

Public Class frmMain
    Private openErrSuccess As Integer = Constants.ME_IDS_NOTOPENED
    Private lineNumber As Integer
    Private subdeviceList As New List(Of FoSubdevice)

    Private Sub frmMain_Shown(sender As System.Object, e As System.EventArgs) Handles MyBase.Shown
        txtBoxFirstPhase.Text = String.Format("{0:G}", 0.2)
        FillPortConfiguration()
        OpenMEiDSDriver()
    End Sub

    Private Sub frmMain_FormClosing(sender As System.Object, e As System.Windows.Forms.FormClosingEventArgs) Handles MyBase.FormClosing
        CloseMEiDSDriver()
    End Sub

    Private Sub btnClose_Click(sender As System.Object, e As System.EventArgs) Handles btnClose.Click
        Close()
    End Sub

    Private Sub OpenMEiDSDriver()
        Try
            textBoxInfo.AppendText("Opening Maindriver..." & Environment.NewLine)
            openErrSuccess = meIDS.meOpen(meIDS.ME_OPEN_NO_FLAGS)

            If openErrSuccess = meIDS.ME_ERRNO_SUCCESS Then
                DisplayVersionLibrary()
                DisplayVersionMainDriver()
                textBoxInfo.AppendText(Environment.NewLine)

                Dim numDevices As Integer
                meIDS.meQueryNumberDevices(numDevices)
                textBoxInfo.AppendText("Number of devices found: " & numDevices.ToString() & Environment.NewLine)

                GetFoSubdevices(numDevices)
            Else
                DisplayErrorMessage("meOpen - Error: ", openErrSuccess)
            End If
        Catch ex As DllNotFoundException
            textBoxInfo.AppendText("The ME-iDS is not properly installed." & Environment.NewLine)
            textBoxInfo.AppendText(ex.Message & Environment.NewLine)
        End Try
    End Sub

    Private Sub GetFoSubdevices(numDevices As Integer)
        If openErrSuccess = meIDS.ME_ERRNO_SUCCESS AndAlso numDevices > 0 Then
            textBoxInfo.AppendText("Retrieving subdevices of type ME_TYPE_FO (Frequency-Out)..." & Environment.NewLine)
            For idxDevice = 0 To numDevices - 1
                Dim meError As Integer
                Dim plugged As Integer
                meError = GetDevicePlugged(idxDevice, plugged)
                If meError = meIDS.ME_ERRNO_SUCCESS AndAlso plugged = meIDS.ME_PLUGGED_IN Then
                    GetSubdevicesByType(idxDevice, meIDS.ME_TYPE_FO)
                Else
                    textBoxInfo.AppendText("Device " & idxDevice.ToString() & " not plugged in" & Environment.NewLine)
                End If
            Next
        End If
        DisplaySubdevices()
    End Sub

    Private Sub GetSubdevicesByType(idxDevice As Integer, subdeviceType As Integer)
        Dim meError As Integer
        Dim numSubdevices As Integer
        Dim idxMatchedSubdevice As Integer

        meError = meIDS.meQueryNumberSubdevices(idxDevice, numSubdevices)
        If meError = meIDS.ME_ERRNO_SUCCESS Then
            Dim deviceName = New StringBuilder(meIDS.ME_DEVICE_NAME_MAX_COUNT)
            meIDS.meQueryNameDevice(idxDevice, deviceName, meIDS.ME_DEVICE_NAME_MAX_COUNT)
            Dim idxSubdevice As Integer = 0
            While idxSubdevice < numSubdevices
                meError = meIDS.meQuerySubdeviceByType(idxDevice, _
                                                 idxSubdevice, _
                                                 subdeviceType, _
                                                 meIDS.ME_SUBTYPE_ANY, _
                                                 idxMatchedSubdevice)

                If meError = meIDS.ME_ERRNO_SUCCESS Then
                    AddFoSubdevice(deviceName.ToString(), idxDevice, idxMatchedSubdevice, subdeviceType)
                    idxSubdevice = idxMatchedSubdevice + 1
                Else
                    Exit While
                End If
            End While
        End If
    End Sub

    Private Sub AddFoSubdevice(deviceName As String, idxDevice As Integer, idxSubdevice As Integer, subdeviceType As Integer)
        Dim numOfChannels As Integer
        Dim devCaps As Integer
        meIDS.meQueryNumberChannels(idxDevice, idxSubdevice, numOfChannels)
        meIDS.meQuerySubdeviceCaps(idxDevice, idxSubdevice, devCaps)
        subdeviceList.Add(New FoSubdevice(deviceName.ToString, idxDevice, idxSubdevice, numOfChannels, devCaps))
        coBoxSubDevs.Items.Add(deviceName.ToString _
                               & ",  Device " & idxDevice.ToString() _
                               & ",  Subdevice " & idxSubdevice.ToString() _
                               & ",  Type: " & GetSubDevTypeName(subdeviceType) _
                               & ",  Channels: " & numOfChannels.ToString())
    End Sub

    Private Sub btnSetConfig_Click(sender As System.Object, e As System.EventArgs) Handles btnSetConfig.Click
        If (coBoxSubDevs.Items.Count > 0) Then
            Dim foSubdev As FoSubdevice = subdeviceList(coBoxSubDevs.SelectedIndex)
            textBoxInfo.AppendText(Environment.NewLine & "Set Config" & Environment.NewLine)

            Dim meError As Integer
            meError = ConfigureFoSubdevice(foSubdev, coBoxChannel.SelectedIndex)
            If (meError = meIDS.ME_ERRNO_SUCCESS) Then
                textBoxInfo.AppendText("Config: Device " + foSubdev.deviceIndex.ToString() _
                                 & ", Subdevice " & foSubdev.subdevIndex.ToString() _
                                 & ", Channel " & coBoxChannel.SelectedIndex.ToString() _
                                 & ", " & GetFoConfigText() _
                                 & Environment.NewLine)
            Else
                DisplayErrorMessage("meIOSingleConfig - Error: ", meError)
            End If
        End If
    End Sub

    Private Function ConfigureFoSubdevice(foSubdevice As FoSubdevice, idxChannel As Integer) As Integer
        Dim singleConfig As Integer = GetSingleConfigParam(foSubdevice)
        Dim triggerChannel As Integer = GetOutputParam()
        Dim meError As Integer = meIDS.meIOSingleConfig(foSubdevice.deviceIndex,
                                    foSubdevice.subdevIndex,
                                    idxChannel,
                                    singleConfig,
                                    meIDS.ME_VALUE_NOT_USED,
                                    triggerChannel,
                                    meIDS.ME_TRIG_TYPE_NONE,
                                    meIDS.ME_VALUE_NOT_USED,
                                    meIDS.ME_IO_SINGLE_CONFIG_NO_FLAGS)

        Return meError
    End Function

    Private Function GetSingleConfigParam(foSubdevice As FoSubdevice) As Integer
        Dim singleConfig As Integer
        If ((foSubdevice.subdeviceCaps And meIDS.ME_CAPS_FIO_SINK_SOURCE) = meIDS.ME_CAPS_FIO_SINK_SOURCE) Then
            singleConfig = coBoxPortConfiguration.SelectedItem.GetValue
        Else
            singleConfig = meIDS.ME_SINGLE_CONFIG_FIO_OUTPUT
        End If
        Return singleConfig
    End Function

    Private Function GetOutputParam() As Integer
        Dim triggerChannel As Integer
        If (rBtnImmediateOutput.Checked) Then
            triggerChannel = meIDS.ME_TRIG_CHAN_DEFAULT
        Else
            triggerChannel = meIDS.ME_TRIG_CHAN_SYNCHRONOUS
        End If
        Return triggerChannel
    End Function

    Private Function GetFoConfigText() As String
        Dim trigChan As String
        If (rBtnImmediateOutput.Checked) Then
            trigChan = rBtnImmediateOutput.Text
        Else
            trigChan = rBtnSynchronousOutput.Text
        End If
        Return trigChan
    End Function

    Private Sub btnReadData_Click(sender As System.Object, e As System.EventArgs) Handles btnReadData.Click
        If (coBoxSubDevs.Items.Count > 0) Then
            Dim foSubdev As FoSubdevice = subdeviceList(coBoxSubDevs.SelectedIndex)
            Dim idxChannel As Integer = coBoxChannel.SelectedIndex
            textBoxInfo.AppendText(Environment.NewLine & "Read Data" & Environment.NewLine)

            Dim valueTotal As Integer
            Dim meError As Integer = ReadSingleValue(foSubdev, idxChannel, meIDS.ME_IO_SINGLE_TYPE_FIO_TICKS_TOTAL, valueTotal)
            If (meError = meIDS.ME_ERRNO_SUCCESS) Then
                Dim value1stPhase As Integer
                meError = ReadSingleValue(foSubdev, idxChannel, meIDS.ME_IO_SINGLE_TYPE_FIO_TICKS_FIRST_PHASE, value1stPhase)
                If (meError = meIDS.ME_ERRNO_SUCCESS) Then
                    DisplayReadBackValues(foSubdev, idxChannel, value1stPhase, valueTotal)
                End If
            End If

            If (meError <> meIDS.ME_ERRNO_SUCCESS) Then
                DisplayErrorMessage("meIOSingle - Error: ", meError)
            End If
        End If
    End Sub

    Private Function ReadSingleValue(foSubdevice As FoSubdevice, idxChannel As Integer, flags As Integer, ByRef value As Integer) As Integer
        Dim io_single(1) As meIDS.meIOSingle_t

        io_single(0).iDevice = foSubdevice.deviceIndex
        io_single(0).iSubdevice = foSubdevice.subdevIndex
        io_single(0).iChannel = idxChannel
        io_single(0).iDir = meIDS.ME_DIR_INPUT
        io_single(0).iValue = 0
        io_single(0).iTimeOut = meIDS.ME_VALUE_NOT_USED
        io_single(0).iFlags = flags

        Dim meError As Integer = meIDS.meIOSingle(io_single(0),
                       1,
                       meIDS.ME_IO_SINGLE_NO_FLAGS)

        value = io_single(0).iValue
        Return meError
    End Function

    Private Sub DisplayReadBackValues(foSubdevice As FoSubdevice, idxChannel As Integer, value1stPhase As Integer, valueTotal As Integer)
        lineNumber = lineNumber + 1
        textBoxInfo.AppendText(String.Format("{0:00}", lineNumber) _
                            & " Read from " & foSubdevice.deviceName _
                            & ",  Device: " & foSubdevice.deviceIndex.ToString() _
                            & ",  Subdevice: " & foSubdevice.subdevIndex.ToString() _
                            & ",  Channel: " & idxChannel.ToString() _
                            & Environment.NewLine)
        textBoxInfo.AppendText("  Ticks total: " & valueTotal.ToString() _
                            & Environment.NewLine)
        textBoxInfo.AppendText("  Ticks first phase: " & value1stPhase.ToString() _
                            & Environment.NewLine)
    End Sub

    Private Sub coBoxSubDevs_SelectedIndexChanged(sender As System.Object, e As System.EventArgs) Handles coBoxSubDevs.SelectedIndexChanged
        coBoxChannel.Items.Clear()

        If (coBoxSubDevs.SelectedIndex >= 0) Then
            Dim foSubdev As FoSubdevice
            foSubdev = subdeviceList(coBoxSubDevs.SelectedIndex)
            Dim i As Integer
            For i = 0 To foSubdev.numOfChannels - 1
                coBoxChannel.Items.Add(i.ToString())
            Next
            coBoxPortConfiguration.Enabled = ((foSubdev.subdeviceCaps And meIDS.ME_CAPS_FIO_SINK_SOURCE) = meIDS.ME_CAPS_FIO_SINK_SOURCE)
        End If

        If (coBoxChannel.Items.Count > 0) Then
            coBoxChannel.SelectedIndex = 0
        End If
    End Sub

    Private Sub FillChannels(foSubdevice As FoSubdevice)
        Dim i As Integer
        For i = 0 To foSubdevice.numOfChannels - 1
            coBoxChannel.Items.Add(i.ToString())
        Next
        If (coBoxChannel.Items.Count > 0) Then
            coBoxChannel.SelectedIndex = 0
        End If
    End Sub

    Private Sub btnWriteData_Click(sender As System.Object, e As System.EventArgs) Handles btnWriteData.Click
        If (coBoxSubDevs.SelectedIndex >= 0) Then
            Dim foSubDev As FoSubdevice = subdeviceList(coBoxSubDevs.SelectedIndex)
            Dim idxChannel As Integer = coBoxChannel.SelectedIndex
            textBoxInfo.AppendText(Environment.NewLine + "Write Data" + Environment.NewLine)

            Try
                Dim secondsAchievedFirstPhase As Double = Double.Parse(txtBoxFirstPhase.Text)
                Dim secondsAchivedTotal As Double = Double.Parse(txtBoxTotalTime.Text)

                Dim ticksTotal As Integer
                Dim meError As Integer = ConvertTimeToTicks(foSubDev, meIDS.ME_TIMER_FIO_TOTAL, secondsAchivedTotal, ticksTotal)
                If (meError = meIDS.ME_ERRNO_SUCCESS) Then
                    Dim flags As Integer = GetFoTotalTimeFlags()
                    meError = WriteTicks(foSubDev, idxChannel, flags, ticksTotal)
                    If (meError = meIDS.ME_ERRNO_SUCCESS) Then
                        Dim ticksFirstPhase As Integer
                        meError = ConvertTimeToTicks(foSubDev, meIDS.ME_TIMER_FIO_FIRST_PHASE, secondsAchievedFirstPhase, ticksFirstPhase)
                        If (meError = meIDS.ME_ERRNO_SUCCESS) Then
                            flags = GetFoFirstPhaseTimeFlags()
                            meError = WriteTicks(foSubDev, idxChannel, flags, ticksFirstPhase)
                            If (meError = meIDS.ME_ERRNO_SUCCESS) Then
                                DisplayTimeValues(foSubDev, idxChannel, secondsAchivedTotal, secondsAchievedFirstPhase, ticksTotal, ticksFirstPhase)
                            End If
                        End If
                    End If
                End If
                If (meError <> meIDS.ME_ERRNO_SUCCESS) Then
                    DisplayErrorMessage("Error: ", meError)
                End If
            Catch ex As System.Exception
                textBoxInfo.AppendText("Error: " & ex.Message)
            End Try
        End If
    End Sub

    Private Function GetFoFirstPhaseTimeFlags() As Integer

        Dim flags As Integer = meIDS.ME_IO_SINGLE_TYPE_FIO_TICKS_FIRST_PHASE
        If rBtnSynchronousOutput.Checked Then
            flags = flags Or meIDS.ME_IO_SINGLE_TYPE_TRIG_SYNCHRONOUS
            If (rBtnStartLevelLow.Checked) Then
                flags = flags Or meIDS.ME_IO_SINGLE_TYPE_FO_START_LOW
            End If

            If (rBtnStarttypeSoft.Checked) Then
                flags = flags Or meIDS.ME_IO_SINGLE_TYPE_FO_START_SOFT
            End If
        End If
        Return flags
    End Function

    Private Function GetFoTotalTimeFlags() As Integer
        Dim flags As Integer = meIDS.ME_IO_SINGLE_TYPE_NO_FLAGS
        If (rBtnSynchronousOutput.Checked) Then
            flags = flags Or meIDS.ME_IO_SINGLE_TYPE_TRIG_SYNCHRONOUS
            flags = flags Or meIDS.ME_IO_SINGLE_TYPE_FIO_TICKS_TOTAL _
                  Or meIDS.ME_IO_SINGLE_TYPE_FO_UPDATE_ONLY
        End If
        Return flags
    End Function

    Private Function ConvertTimeToTicks(foSubdevice As FoSubdevice, timer As Integer, ByRef secondsAchived As Double, ByRef ticks As Integer) As Integer
        Dim ticksHigh As Integer
        Dim meError As Integer = meIDS.meIOSingleTimeToTicks(foSubdevice.deviceIndex,
                foSubdevice.subdevIndex,
                timer,
                secondsAchived,
                ticks,
                ticksHigh,
                meIDS.ME_IO_SINGLE_TIME_TO_TICKS_NO_FLAGS)

        Return meError
    End Function

    Private Function WriteTicks(foSubdevice As FoSubdevice, idxChannel As Integer, flags As Integer, ticks As Integer) As Integer
        Dim io_single(1) As meIDS.meIOSingle_t

        io_single(0).iDevice = foSubdevice.deviceIndex
        io_single(0).iSubdevice = foSubdevice.subdevIndex
        io_single(0).iChannel = idxChannel
        io_single(0).iDir = meIDS.ME_DIR_OUTPUT
        io_single(0).iValue = ticks
        io_single(0).iTimeOut = 0
        io_single(0).iFlags = flags

        Dim meError As Integer = meIDS.meIOSingle(io_single(0),
                                           1,
                                           meIDS.ME_IO_SINGLE_NO_FLAGS)
        Return meError
    End Function

    Private Sub DisplayTimeValues(foSubdevice As FoSubdevice, idxChannel As Integer, secondsTotal As Double, secondsFirstPhase As Double, ticksTotal As Integer, ticksFirstPhase As Integer)
        lineNumber = lineNumber + 1
        textBoxInfo.AppendText(String.Format("{0:00}", lineNumber) _
                           & " Write to " & foSubdevice.deviceName _
                           & ",  Device: " & foSubdevice.deviceIndex.ToString() _
                           & ",  Subdevice: " & foSubdevice.subdevIndex.ToString() _
                           & ",  Channel: " & idxChannel.ToString() _
                           & Environment.NewLine)

        textBoxInfo.AppendText(String.Format("  Total time: {0:0.##########} sec, frequency: {1:0.##########} Hz", secondsTotal, 1 / secondsTotal) _
                           & ",  Ticks: " & ticksTotal.ToString() _
                           & Environment.NewLine)

        textBoxInfo.AppendText(String.Format("  First phase time: {0:0.##########} sec", secondsFirstPhase) _
                           & ",  Ticks: " & ticksFirstPhase.ToString() _
                           & Environment.NewLine)
    End Sub

    Private Sub btnResetSubdev_Click(sender As System.Object, e As System.EventArgs) Handles btnResetSubdev.Click
        Dim foSubdev As FoSubdevice
        Dim meError As Integer

        foSubdev = subdeviceList(coBoxSubDevs.SelectedIndex)
        meError = meIDS.meIOResetSubdevice(foSubdev.deviceIndex, _
                                     foSubdev.subdevIndex, _
                                     meIDS.ME_IO_RESET_SUBDEVICE_NO_FLAGS)
        If meError = meIDS.ME_ERRNO_SUCCESS Then
            textBoxInfo.AppendText("Reset Device " & foSubdev.deviceIndex.ToString() & ", Subdevice " & foSubdev.subdevIndex.ToString() & Environment.NewLine)
        Else
            DisplayErrorMessage("Error resetting Subdevice ", meError)
        End If
    End Sub

    Private Sub CloseMEiDSDriver()
        If (openErrSuccess = meIDS.ME_ERRNO_SUCCESS) Then
            meIDS.meClose(meIDS.ME_CLOSE_NO_FLAGS)
            openErrSuccess = Constants.ME_IDS_NOTOPENED
            btnReadData.Enabled = False
            btnResetSubdev.Enabled = False
            btnSetConfig.Enabled = False
            subdeviceList.Clear()
            coBoxSubDevs.Items.Clear()
            coBoxChannel.Items.Clear()
        End If
    End Sub

    Private Sub DisplaySubdevices()
        If (coBoxSubDevs.Items.Count > 0) Then
            textBoxInfo.AppendText("Number of Frequency-Out subdevices found: " & coBoxSubDevs.Items.Count.ToString() & Environment.NewLine)
            For i = 0 To coBoxSubDevs.Items.Count - 1
                textBoxInfo.AppendText(coBoxSubDevs.Items(i) & Environment.NewLine)
            Next
            textBoxInfo.AppendText(Environment.NewLine)
            coBoxSubDevs.SelectedIndex = 0
            btnReadData.Enabled = True
            btnResetSubdev.Enabled = True
            btnWriteData.Enabled = True
            btnSetConfig.Enabled = True
        Else
            textBoxInfo.AppendText("No Frequency-Out subdevices found." & Environment.NewLine)
            meIDS.meClose(meIDS.ME_CLOSE_NO_FLAGS)
            openErrSuccess = Constants.ME_IDS_NOTOPENED
        End If
    End Sub

    Private Sub FillPortConfiguration()
        coBoxPortConfiguration.Items.Add(New ComboBoxItem("High impedance", meIDS.ME_SINGLE_CONFIG_FIO_HIGH_IMPEDANCE))
        coBoxPortConfiguration.Items.Add(New ComboBoxItem("Sink", meIDS.ME_SINGLE_CONFIG_FIO_SINK))
        coBoxPortConfiguration.Items.Add(New ComboBoxItem("Source", meIDS.ME_SINGLE_CONFIG_FIO_SOURCE))
        coBoxPortConfiguration.SelectedIndex = 0
    End Sub

    Private Function GetSubDevTypeName(subdeviceType As Integer)
        Dim tmpType As String
        Select Case subdeviceType
            Case meIDS.ME_TYPE_INVALID
                tmpType = "ME_TYPE_INVALID"
            Case meIDS.ME_TYPE_AI
                tmpType = "ME_TYPE_AI"
            Case meIDS.ME_TYPE_AO
                tmpType = "ME_TYPE_AO"
            Case meIDS.ME_TYPE_DIO
                tmpType = "ME_TYPE_DIO"
            Case meIDS.ME_TYPE_DO
                tmpType = "ME_TYPE_DO"
            Case meIDS.ME_TYPE_DI
                tmpType = "ME_TYPE_DI"
            Case meIDS.ME_TYPE_CTR
                tmpType = "ME_TYPE_CTR"
            Case meIDS.ME_TYPE_EXT_IRQ
                tmpType = "ME_TYPE_EXT_IRQ"
            Case meIDS.ME_TYPE_FIO
                tmpType = "ME_TYPE_FIO"
            Case meIDS.ME_TYPE_FO
                tmpType = "ME_TYPE_FO"
            Case meIDS.ME_TYPE_FI
                tmpType = "ME_TYPE_FI"
            Case meIDS.ME_TYPE_FPGA
                tmpType = "ME_TYPE_FPGA"
            Case Else
                tmpType = String.Format("&H{0:X}", subdeviceType)
        End Select
        Return tmpType
    End Function

    Private Sub DisplayErrorMessage(message As String, meError As Integer)
        Dim errMessage = New StringBuilder(meIDS.ME_ERROR_MSG_MAX_COUNT)
        textBoxInfo.AppendText(message + meError.ToString() & Environment.NewLine)
        meIDS.meErrorGetMessage(meError, errMessage, meIDS.ME_ERROR_MSG_MAX_COUNT)
        textBoxInfo.AppendText("ErrorMessage: " & errMessage.ToString & Environment.NewLine)
    End Sub

    Private Sub DisplayVersionLibrary()
        Dim tmpVersion As Integer
        meIDS.meQueryVersionLibrary(tmpVersion)
        textBoxInfo.AppendText("Library version: " & MeVersionToString(tmpVersion) & Environment.NewLine)
    End Sub

    Private Sub DisplayVersionMainDriver()
        Dim tmpVersion As Integer
        meIDS.meQueryVersionMainDriver(tmpVersion)
        textBoxInfo.AppendText("Maindriver version: " & MeVersionToString(tmpVersion) & Environment.NewLine)
    End Sub

    Private Function GetDevicePlugged(idxDevice As Integer, ByRef plugged As Integer) As Integer

        Dim meError As Integer
        Dim vendorId As Integer
        Dim deviceId As Integer
        Dim serialNo As Integer
        Dim busType As Integer
        Dim busNo As Integer
        Dim devNo As Integer
        Dim funcNo As Integer

        meError = meIDS.meQueryInfoDevice(idxDevice, _
                                    vendorId, _
                                    deviceId, _
                                    serialNo, _
                                    busType, _
                                    busNo, _
                                    devNo, _
                                    funcNo, _
                                    plugged)
        Return meError
    End Function

    Private Function MeVersionToString(version As Integer) As String
        Return String.Format("{0}.{1}.{2}.{3}", version >> 24, &HFF And (version >> 16), &HFF And (version >> 8), &HFF And version)
    End Function

    Private Sub btnClearHwInfo_Click(sender As System.Object, e As System.EventArgs) Handles btnClearHwInfo.Click
        textBoxInfo.Clear()
        lineNumber = 0
    End Sub

    Private Sub btnAbout_Click(sender As System.Object, e As System.EventArgs) Handles btnAbout.Click
        AboutBox1.ShowDialog()
    End Sub
End Class


Public Class FoSubdevice
    Public deviceName As String
    Public deviceIndex As Integer
    Public subdevIndex As Integer
    Public numOfChannels As Integer
    Public subdeviceCaps As Integer
    Sub New(DevName As String, DeviceIdx As Integer, SubdevIdx As Integer, NumChannels As Integer, SubdevCaps As Integer)
        deviceName = DevName
        deviceIndex = DeviceIdx
        subdevIndex = SubdevIdx
        numOfChannels = NumChannels
        subdeviceCaps = SubdevCaps
    End Sub
End Class

Public Class ComboBoxItem
    Dim dispValue As String
    Dim val As Integer

    Sub New(displayValue As String, value As Integer)
        dispValue = displayValue
        val = value
    End Sub

    Public Overrides Function ToString() As String
        Return dispValue
    End Function

    ReadOnly Property GetValue() As Integer
        Get
            Return val
        End Get
    End Property
End Class

Class Constants
    Public Shared ReadOnly ME_IDS_NOTOPENED As Integer = -1
End Class