﻿//==============================================================================
//Copyright(c), Meilhaus Electronic GmbH
//              http://www.meilhaus.de
//              support@meilhaus.de
//
//Source File : FormMain.cs
//Description : Example program for reading values from Frequency-In (FI) subdevices
//
//==============================================================================
//  Note about the hardware:
//  After meIOSingle is called (start of the measurement operation) 
//  the FI hardware will automatically detect the next rising edge 
//  of the unknown digital signal to start the measurement of the signal period.
//  The first counter (counter A) on the hardware is started.
//  Counter A is stopped when the next falling edge is detected.
//  At the same time a second counter (counter B) is started that will stop on
//  detection of the next rising edge (the end of the signal period).
//  The two counter values describe the signal period
//  The ME-iDS function library will provide two values
//  1. the counter value for the entire signal period (counter A + counter B)
//  2. the counter value of the 1st phase of the signal period (counter A)
//
//  Reading Data:
//  A new measurement operation is started with the given timeout value
//  (iFlag is set to ME_IO_SINGLE_CONFIG_NO_FLAGS)
//  After the measuring is done (meIOSingle() returned successfully)
//  the value of Total ticks is returned in member iValue of meIOSingle_t,
//  Then meIOSingle is called a second time to read the value of the first phase.
//  Therefore ME_IO_SINGLE_TYPE_FI_LAST_VALUE must be specified in iFlags
//  (If ME_IO_SINGLE_TYPE_FI_LAST_VALUE isn't specified a new measuring
//  would be started and the returned counter value wouldn't belong
//  to the same signal period, thus giving wrong results)
//  With the function meIOSingleTicksToTime() these values can be converted to a time value.
//
//  unknown digital signal:
//      +-------+                   +---
//      |       |                   |   
//      |       |                   |   
//  ----+       +-------------------+   
//                                   
//  ^---^-------^-------------------^
//  0   1       2                   3
//
//  0: Start of the operation (call of function meIOSingle())
//  1: Start of hardware counter A on rising edge of the signal
//  2: End of the first phase of the signal, stop of hardware counter A,
//     Start of hardware counter B
//  3: End of the second phase of the signal, stop of hardware counter B
//     End of the operation (function meIOSingle() returned)
//

//------------------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using meIDS_CS;

namespace Single_Read_Frequency
{
    public partial class FormMain : Form
    {
        private int openErrSuccess = Constants.ME_IDS_NOTOPENED;
        private int lineNumber = 0;
        private List<FiSubdevice> subdeviceList = new List<FiSubdevice>();      

        public FormMain()
        {
            InitializeComponent();
        }

        private void FormMain_Shown(object sender, EventArgs e)
        {
            OpenMEiDSDriver();
        }

        private void FormMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            CloseMEiDSDriver();
        }

        private void btnClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void OpenMEiDSDriver()
        {
            try
            {
                textBoxInfo.AppendText("Opening Maindriver..." + Environment.NewLine);
                openErrSuccess = meIDS.meOpen(meIDS.ME_OPEN_NO_FLAGS);          
    
                if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS)
                {
                    DisplayVersionLibrary();
                    DisplayVersionMainDriver();
                    textBoxInfo.AppendText(Environment.NewLine);
    
                    int numDevices;
                    meIDS.meQueryNumberDevices(out numDevices);                     
                    textBoxInfo.AppendText("Number of devices found: " + numDevices.ToString() + Environment.NewLine);
    
                    GetFiSubdevices(numDevices);
                }
                else
                {
                    DisplayErrorMessage("meOpen - Error: ", openErrSuccess);
                }
            }
            catch (DllNotFoundException ex)
            {
                textBoxInfo.AppendText("The ME-iDS is not properly installed." + Environment.NewLine);
                textBoxInfo.AppendText(ex.Message + Environment.NewLine);
            }
        }

        private void GetFiSubdevices(int numDevices)
        {
            if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS && numDevices > 0)
            {
                textBoxInfo.AppendText("Retrieving subdevices of type ME_TYPE_FI (Frequency-In)..." + Environment.NewLine);
                for (int idxDevice = 0; idxDevice < numDevices; idxDevice++)
                {
                    int plugged;
                    int meError = GetDevicePlugged(idxDevice, out plugged);

                    if (meError == meIDS.ME_ERRNO_SUCCESS && plugged == meIDS.ME_PLUGGED_IN)
                    {
                        GetSubdevicesByType(idxDevice, meIDS.ME_TYPE_FI);  
                    }
                    else
                    {
                        textBoxInfo.AppendText("Device " + idxDevice.ToString() + " not plugged in" + Environment.NewLine);
                    }
                }
            }
            DisplaySubdevices();
        }

        private void GetSubdevicesByType(int idxDevice, int subdeviceType)
        {
            int meError;
            int numSubdevices;
            int idxMatchedSubdevice;
            StringBuilder devName = new StringBuilder(meIDS.ME_DEVICE_NAME_MAX_COUNT);
            
            meError = meIDS.meQueryNumberSubdevices(idxDevice, out numSubdevices);  
            if (meError == meIDS.ME_ERRNO_SUCCESS)
            {
                meIDS.meQueryNameDevice(idxDevice, devName, devName.Capacity);  
                int idxSubdevice = 0;
                while (idxSubdevice < numSubdevices)
                {
                    meError = meIDS.meQuerySubdeviceByType(idxDevice,                
                                                           idxSubdevice,             
                                                           subdeviceType,            
                                                           meIDS.ME_SUBTYPE_ANY,     
                                                           out idxMatchedSubdevice); 

                    if (meError == meIDS.ME_ERRNO_SUCCESS)
                    {
                        AddFiSubdevice(devName.ToString(), idxDevice, idxMatchedSubdevice, subdeviceType);                       
                        idxSubdevice = idxMatchedSubdevice + 1;                 
                    }
                    else
                    {
                        break;                                                  
                    }
                }
            }
        }

        private void AddFiSubdevice(String devName, int idxDevice, int idxSubdevice, int subdeviceType)
        {
            int numOfChannels;
            meIDS.meQueryNumberChannels(idxDevice, idxSubdevice, out numOfChannels);   
            subdeviceList.Add(new FiSubdevice(devName, idxDevice, idxSubdevice, numOfChannels));   
            coBoxSubDevs.Items.Add(devName
                                    + ",  Device " + idxDevice.ToString()
                                    + ",  Subdevice " + idxSubdevice.ToString()
                                    + ",  Type: " + GetSubDevTypeName(subdeviceType)
                                    + ",  Channels: " + numOfChannels.ToString());
        }

        private void CloseMEiDSDriver()
        {
            if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS)
            {
                meIDS.meClose(meIDS.ME_CLOSE_NO_FLAGS);                         
                openErrSuccess = Constants.ME_IDS_NOTOPENED;
                btnSingleRead.Enabled = false;
                btnResetSubdev.Enabled = false;
                subdeviceList.Clear();
                coBoxSubDevs.Items.Clear();
            }
        }

        private void btnResetSubdev_Click(object sender, EventArgs e)
        {
            FiSubdevice fiSubdev = subdeviceList[coBoxSubDevs.SelectedIndex];
            int meError = meIDS.meIOResetSubdevice(fiSubdev.deviceIndex,
                                                  fiSubdev.subdevIndex,
                                                  meIDS.ME_IO_RESET_SUBDEVICE_NO_FLAGS);
            if (meError == meIDS.ME_ERRNO_SUCCESS)
            {
                textBoxInfo.AppendText("Reset Device " + fiSubdev.deviceIndex.ToString() + ", Subdevice " + fiSubdev.subdevIndex.ToString() + Environment.NewLine);
            }
            else
            {
                DisplayErrorMessage("Error resetting Subdevice", meError);
            }
        }

        private void btnSingleRead_Click(object sender, EventArgs e)
        {
            if (coBoxSubDevs.SelectedIndex >= 0)
            {
                EnableUI(false);
                try
                {
                    FiSubdevice fiSubdevice = subdeviceList[coBoxSubDevs.SelectedIndex];
                    int idxChannel = coBoxChannel.SelectedIndex;
                    int timeout = (int)numUpDownTimeout.Value;
                    
                    int meError = ConfigureSubdeviceForInput(fiSubdevice, idxChannel);
                    if (meError == meIDS.ME_ERRNO_SUCCESS)
                    {
                        int valueTotal;
                        int flags = meIDS.ME_IO_SINGLE_TYPE_FIO_TICKS_TOTAL;
                        meError = ReadSingleValue(fiSubdevice, idxChannel, timeout, flags, out valueTotal);
                        if (meError == meIDS.ME_ERRNO_SUCCESS)
                        {
                            int value1stPhase;
                            flags = meIDS.ME_IO_SINGLE_TYPE_FI_LAST_VALUE | meIDS.ME_IO_SINGLE_TYPE_FIO_TICKS_FIRST_PHASE;
                            
                            meError = ReadSingleValue(fiSubdevice, idxChannel, timeout, flags, out value1stPhase);
                            if (meError == meIDS.ME_ERRNO_SUCCESS)
                            {
                                meError = DisplayValues(fiSubdevice, idxChannel, valueTotal, value1stPhase);                            
                            }
                        }
                    }
                    if (meError != meIDS.ME_ERRNO_SUCCESS)
                    {
                        DisplayErrorMessage("Error: ", meError);                    
                    }
                }
                finally
                {
                    EnableUI(true);
                }
            }
        }

        private int ReadSingleValue(FiSubdevice fiSubdevice, int idxChannel, int timeout, int flags, out int value)
        {
            meIDS.meIOSingle_t[] io_single = new meIDS.meIOSingle_t[1];     
            
            io_single[0].iDevice = fiSubdevice.deviceIndex;                 
            io_single[0].iSubdevice = fiSubdevice.subdevIndex;              
            io_single[0].iChannel = idxChannel;                             
            io_single[0].iDir = meIDS.ME_DIR_INPUT;                        
            io_single[0].iValue = 0;                                        
            io_single[0].iTimeOut = timeout;                                
            io_single[0].iFlags = flags;                                    

            int meError = meIDS.meIOSingle(ref io_single[0],                
                                    1,                                      
                                    meIDS.ME_IO_SINGLE_NO_FLAGS);
            value = io_single[0].iValue;
            return meError;
        }

        private int DisplayValues(FiSubdevice fiSubdevice, int idxChannel, int valueTotal, int value1stPhase)
        {
            Double firstPhaseTime = 0;
            Double totalTime = 0;
            
            int meError = CalculateTime(fiSubdevice, valueTotal, meIDS.ME_TIMER_FIO_TOTAL, ref totalTime);
            if (meError == meIDS.ME_ERRNO_SUCCESS)
            {
                CalculateTime(fiSubdevice, value1stPhase, meIDS.ME_TIMER_FIO_FIRST_PHASE, ref firstPhaseTime);
                lineNumber++;
                textBoxInfo.AppendText(String.Format("{0:00}", lineNumber)     
                                        + " Read from " + fiSubdevice.deviceName
                                        + ",  Device: " + fiSubdevice.deviceIndex.ToString()
                                        + ",  Subdevice: " + fiSubdevice.subdevIndex.ToString()
                                        + ",  Channel: " + idxChannel.ToString());
                textBoxInfo.AppendText("  Ticks total: " + valueTotal.ToString()
                                        + String.Format(",  Seconds: {0:0.##########},  Hz: {1:0.##########}", totalTime, 1 / totalTime)); 
                textBoxInfo.AppendText("  Ticks 1st phase: " + value1stPhase.ToString()
                                        + String.Format(",  Seconds: {0:0.##########},  Duty Cycle: {1:0.###}%", firstPhaseTime, (valueTotal / value1stPhase) * 100)); 
            }
            return meError;
        }

        private int CalculateTime(FiSubdevice fiSubdevice, int valueDigital, int timer, ref Double time)
        {
            int meError = meIDS.meIOSingleTicksToTime(fiSubdevice.deviceIndex,
                                  fiSubdevice.subdevIndex,
                                  timer, 
                                  valueDigital,
                                  0,
                                  ref time,
                                  meIDS.ME_IO_SINGLE_TICKS_TO_TIME_NO_FLAGS);

            return meError;
        }


        private void btnConfig_Click(object sender, EventArgs e)
        {
        }

        private static int ConfigureSubdeviceForInput(FiSubdevice fiSubdevice, int idxChannel)
        {
            int meError = meIDS.meIOSingleConfig(fiSubdevice.deviceIndex,
                                        fiSubdevice.subdevIndex,
                                        idxChannel,
                                        meIDS.ME_SINGLE_CONFIG_FIO_INPUT,
                                        meIDS.ME_VALUE_NOT_USED,
                                        meIDS.ME_TRIG_CHAN_DEFAULT,
                                        meIDS.ME_TRIG_TYPE_NONE,
                                        meIDS.ME_VALUE_NOT_USED,
                                        meIDS.ME_IO_SINGLE_CONFIG_NO_FLAGS);
            return meError;
        }

        private void coBoxSubDevs_SelectedIndexChanged(object sender, EventArgs e)
        {
            coBoxChannel.Items.Clear();
            if (coBoxSubDevs.SelectedIndex >= 0)
            {
                FiSubdevice fiSubdevice;
                fiSubdevice = subdeviceList[coBoxSubDevs.SelectedIndex];
                for (int i = 0; i < fiSubdevice.numOfChannels; i++)
                {
                    coBoxChannel.Items.Add(i.ToString());
                }
            }
            if (coBoxChannel.Items.Count > 0) coBoxChannel.SelectedIndex = 0;
        }

        private void DisplaySubdevices()
        {
            if (coBoxSubDevs.Items.Count > 0)
            {
                textBoxInfo.AppendText("Number of Frequency-In subdevices found: " + coBoxSubDevs.Items.Count.ToString() + Environment.NewLine);
                for (int i = 0; i < coBoxSubDevs.Items.Count; i++)
                {
                    textBoxInfo.AppendText(coBoxSubDevs.Items[i] + Environment.NewLine);
                }
                textBoxInfo.AppendText(Environment.NewLine);
                coBoxSubDevs.SelectedIndex = 0;
                btnSingleRead.Enabled = true;
                btnResetSubdev.Enabled = true;
            }
            else
            {
                textBoxInfo.AppendText("No Frequency-In subdevices found." + Environment.NewLine);
                CloseMEiDSDriver();
            }
        }

        private int GetDevicePlugged(int idxDevice, out int plugged)
        {
            int vendorId;
            int deviceId;
            int serialNo;
            int busType;
            int busNo;
            int devNo;
            int funcNo;

            int meError = meIDS.meQueryInfoDevice(idxDevice,
                                             out vendorId,
                                             out deviceId,
                                             out serialNo,
                                             out busType,
                                             out busNo,
                                             out devNo,
                                             out funcNo,
                                             out plugged);
            return meError;
        }

        private void DisplayVersionMainDriver()
        {
            int tmpVersion;
            meIDS.meQueryVersionMainDriver(out tmpVersion);
            textBoxInfo.AppendText("Maindriver version: " + MeVersionToString(tmpVersion) + Environment.NewLine);
        }

        private void DisplayVersionLibrary()
        {
            int tmpVersion;
            meIDS.meQueryVersionLibrary(out tmpVersion);
            textBoxInfo.AppendText("Library version: " + MeVersionToString(tmpVersion) + Environment.NewLine);
        }

        private string GetSubDevTypeName(int subdeviceType)
        {
            string tmpType;
            switch (subdeviceType)
            {
                case meIDS.ME_TYPE_INVALID:
                    tmpType = "ME_TYPE_INVALID";
                    break;
                case meIDS.ME_TYPE_AI:
                    tmpType = "ME_TYPE_AI";
                    break;
                case meIDS.ME_TYPE_AO:
                    tmpType = "ME_TYPE_AO";
                    break;
                case meIDS.ME_TYPE_DIO:
                    tmpType = "ME_TYPE_DIO";
                    break;
                case meIDS.ME_TYPE_DO:
                    tmpType = "ME_TYPE_DO";
                    break;
                case meIDS.ME_TYPE_DI:
                    tmpType = "ME_TYPE_DI";
                    break;
                case meIDS.ME_TYPE_CTR:
                    tmpType = "ME_TYPE_CTR";
                    break;
                case meIDS.ME_TYPE_EXT_IRQ:
                    tmpType = "ME_TYPE_EXT_IRQ";
                    break;
                case meIDS.ME_TYPE_FIO:
                    tmpType = "ME_TYPE_FIO";
                    break;
                case meIDS.ME_TYPE_FO:
                    tmpType = "ME_TYPE_FO";
                    break;
                case meIDS.ME_TYPE_FI:
                    tmpType = "ME_TYPE_FI";
                    break;
                case meIDS.ME_TYPE_FPGA:
                    tmpType = "ME_TYPE_FPGA";
                    break;
                default:
                    tmpType = String.Format("0x{0:X}", subdeviceType);
                    break;
            };
            return tmpType;
        }

        private void DisplayErrorMessage(string message, int meError)
        {
            StringBuilder errMessage = new StringBuilder(meIDS.ME_ERROR_MSG_MAX_COUNT);
            textBoxInfo.AppendText(message + meError.ToString() + Environment.NewLine);
            meIDS.meErrorGetMessage(meError, errMessage, errMessage.Capacity);             
            textBoxInfo.AppendText("ErrorMessage: " + errMessage + Environment.NewLine);    
        }

        private void EnableUI(bool isEnabled)
        {
            coBoxSubDevs.Enabled = isEnabled;
            coBoxChannel.Enabled = isEnabled;
            btnSingleRead.Enabled = isEnabled;
            numUpDownTimeout.Enabled = isEnabled;
        }

        private string MeVersionToString(int version)
        {
            return String.Format("{0}.{1}.{2}.{3}", version >> 24, 0xFF & (version >> 16), 0xFF & (version >> 8), 0xFF & version);
        }

        private void btnClearHwInfo_Click(object sender, EventArgs e)
        {
            textBoxInfo.Clear();
            lineNumber = 0;
        }

        private void btnAbout_Click(object sender, EventArgs e)
        {
            AboutBox1 aboutBox = new AboutBox1();
            aboutBox.ShowDialog();
        }
    }

    class FiSubdevice
    {
        public string deviceName;
        public int deviceIndex;
        public int subdevIndex;
        public int numOfChannels;
        public FiSubdevice(string DeviceName, int DeviceIndex, int SubdevIndex, int NumOfChannels)
        {
            this.deviceName = DeviceName;
            this.deviceIndex = DeviceIndex;
            this.subdevIndex = SubdevIndex;
            this.numOfChannels = NumOfChannels;
        }
    }

    static class Constants
    {
        public const int ME_IDS_NOTOPENED = -1;
    }
}
