﻿//==============================================================================
//Copyright(c), Meilhaus Electronic GmbH
//              http://www.meilhaus.de
//              support@meilhaus.de
//
//Source File : FormMain.cs
//Description : Example program for writing single values to Digital-out (DO) and Digital-in/out (DIO) devices
//              - Open ME-iDS main driver when program is started
//              - list all ME-iDS devices found in system
//              - determine subdevices of DO and DIO type
//              - Add DO and DIO subdevices to the subdevice combobox
//              - Write data to selected subdevice when user clicks on button "Write Value"
//              - Data are always written to channel 0 with full channel width in this example
//              - Close ME-iDS main driver when program is closed
//==============================================================================
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using meIDS_CS;

namespace Single_Write_Digital
{
    public partial class frmMain : Form
    {
        private int openErrSuccess = Constants.ME_IDS_NOTOPENED;
        private int dataLine = 0;
        private List<DoSubdevice> subdeviceList = new List<DoSubdevice>();      

        public frmMain()
        {
            InitializeComponent();
        }

        private void frmMain_Shown(object sender, EventArgs e)
        {
            FillPortConfiguration();
            OpenMEiDSDriver();
        }

        private void frmMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            CloseMEiDSDriver();
        }

        private void btnClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void OpenMEiDSDriver()
        {
            try
            {
                textBoxInfo.AppendText("Opening Maindriver..." + Environment.NewLine);
                openErrSuccess = meIDS.meOpen(meIDS.ME_OPEN_NO_FLAGS);          

                if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS)
                {
                    DisplayVersionLibrary();
                    DisplayVersionMainDriver();
                    textBoxInfo.AppendText(Environment.NewLine);

                    int numDevices;
                    meIDS.meQueryNumberDevices(out numDevices);                     
                    textBoxInfo.AppendText("Number of devices found: " + (numDevices).ToString() + Environment.NewLine);

                    GetDoDioSubdevices(numDevices);
                }
                else
                {
                    DisplayErrorMessage("meOpen - Error: ", openErrSuccess);
                }
            }
            catch (DllNotFoundException ex)
            {
                textBoxInfo.AppendText("The ME-iDS is not properly installed." + Environment.NewLine);
                textBoxInfo.AppendText(ex.Message + Environment.NewLine);
            }
        }

        private void GetDoDioSubdevices(int numDevices)
        {
            if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS && numDevices > 0)
            {
                textBoxInfo.AppendText("Retrieving sub-devices of type ME_TYPE_DO (Digital Output) and ME_TYPE_DIO (Digital In/Out)..." + Environment.NewLine);
                for (int idxDevice = 0; idxDevice < numDevices; idxDevice++)
                {
                    int plugged;
                    int meError = GetDevicePlugged(idxDevice, out plugged);

                    if (meError == meIDS.ME_ERRNO_SUCCESS && plugged == meIDS.ME_PLUGGED_IN)
                    {
                        GetSubdevicesByType(idxDevice, meIDS.ME_TYPE_DO);  
                        GetSubdevicesByType(idxDevice, meIDS.ME_TYPE_DIO); 
                    }
                    else
                    {
                        textBoxInfo.AppendText("Device index " + idxDevice.ToString() + " not plugged in" + Environment.NewLine);
                    }
                }
            }
            DisplaySubdevices();
        }

        private void DisplaySubdevices()
        {
            if (coBoxSubDevs.Items.Count > 0)
            {
                textBoxInfo.AppendText("Number of DO or DIO sub-devices found: " + coBoxSubDevs.Items.Count.ToString() + Environment.NewLine);
                for (int i = 0; i < coBoxSubDevs.Items.Count; i++)
                {
                    textBoxInfo.AppendText(coBoxSubDevs.Items[i] + Environment.NewLine);  
                }
                
                textBoxInfo.AppendText(Environment.NewLine);
                coBoxSubDevs.SelectedIndex = 0;                             
                btnSingleWrite.Enabled = true;
                btnWriteZero.Enabled = true;
                btnResetSubdev.Enabled = true;
            }
            else
            {
                textBoxInfo.AppendText("No DO or DIO sub-devices found." + Environment.NewLine);
                CloseMEiDSDriver();
            }
        }

        private void GetSubdevicesByType(int idxDevice, int subdeviceType)
        {
            int meError;
            int numSubdevices;
            int idxMatchedSubdevice;
            
            meError = meIDS.meQueryNumberSubdevices(idxDevice, out numSubdevices);   
            if (meError == meIDS.ME_ERRNO_SUCCESS)
            {
                int idxSubdevice = 0;
                StringBuilder devName = new StringBuilder(meIDS.ME_DEVICE_NAME_MAX_COUNT);
                meIDS.meQueryNameDevice(idxDevice, devName, devName.Capacity);  
                while (idxSubdevice < numSubdevices)
                {
                    meError = meIDS.meQuerySubdeviceByType(idxDevice,                 
                                                           idxSubdevice,              
                                                           subdeviceType,             
                                                           meIDS.ME_SUBTYPE_ANY,      
                                                           out idxMatchedSubdevice); 

                    if (meError == meIDS.ME_ERRNO_SUCCESS)
                    {
                        AddDoSubdevice(devName.ToString(), idxDevice, idxMatchedSubdevice, subdeviceType);                        
                        idxSubdevice = idxMatchedSubdevice + 1;                 
                    }
                    else
                    {
                        break;                                                  
                    }
                }
            }
        }

        private void AddDoSubdevice(String devName, int idxDevice, int idxMatchedSubdevice, int subdeviceType)
        {
            int numOfChannels;
            int devCaps;

            meIDS.meQueryNumberChannels(idxDevice, idxMatchedSubdevice, out numOfChannels);   
            meIDS.meQuerySubdeviceCaps(idxDevice, idxMatchedSubdevice, out devCaps);      
            subdeviceList.Add(new DoSubdevice(devName, idxDevice, idxMatchedSubdevice, numOfChannels, devCaps));   

            coBoxSubDevs.Items.Add(devName
                                   + ",  Device " + idxDevice.ToString()
                                   + ",  Subdevice " + idxMatchedSubdevice.ToString()
                                   + ",  Type: " + GetSubDevTypeName(subdeviceType)
                                   + ",  Channels: " + numOfChannels.ToString());
        }

        private void coBoxSubDevs_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (coBoxSubDevs.SelectedIndex >= 0)
            {
                DoSubdevice doSubdev = subdeviceList[coBoxSubDevs.SelectedIndex];
                coBoxPortConfiguration.Enabled = ((doSubdev.deviceCaps & meIDS.ME_CAPS_DIO_SINK_SOURCE) == meIDS.ME_CAPS_DIO_SINK_SOURCE);
            }
            else
            {
                coBoxPortConfiguration.Enabled = false;
            }
        }


        private void btnSingleWrite_Click(object sender, EventArgs e)
        {
            if (coBoxSubDevs.SelectedIndex >= 0)
            {
                DoSubdevice doSubdev = subdeviceList[coBoxSubDevs.SelectedIndex];
                int valueToWrite;
                try
                {
                    valueToWrite = CheckValueToWrite(txtBoxHexVal.Text, doSubdev.numOfChannels);
                    WriteValue(doSubdev, valueToWrite);
                }
                catch (System.Exception ex)
                {
                    MessageBox.Show("Error: " + ex.Message);
                    txtBoxHexVal.Focus();
                }
            }
        }

        private int CheckValueToWrite(string valStr, int numChannels)
        {
            uint x = uint.Parse(valStr, System.Globalization.NumberStyles.AllowHexSpecifier);
            if (numChannels < 32)
            {
                int m = (1 << numChannels) -1;
                if(x > m) {
                    throw new Exception(String.Format("The value is out of range. Maximum value Hex: {0:X}", m));
                }
            }
            return (int)x;
        }

        private void btnWriteZero_Click(object sender, EventArgs e)
        {
            if (coBoxSubDevs.SelectedIndex >= 0)
            {
                DoSubdevice doSubdev = subdeviceList[coBoxSubDevs.SelectedIndex];
                WriteValue(doSubdev, 0);
            }
        }

        private void WriteValue(DoSubdevice doSubdev, int valueToWrite)
        {
            int meError = ConfigureSubdeviceForOutput(doSubdev);
            if (meError == meIDS.ME_ERRNO_SUCCESS)
            {
                meError = WriteSingleValue(doSubdev, valueToWrite);
                if (meError == meIDS.ME_ERRNO_SUCCESS)
                {
                    DisplayValue(doSubdev, valueToWrite);
                }
                else
                {
                    DisplayErrorMessage("meIOSingle - Error: ", meError);
                }
            }
            else
            {
                DisplayErrorMessage("meIOSingleConfig - Error: ", meError);
            }
        }

        private void DisplayValue(DoSubdevice doSubdev, int valueToWrite)
        {
            dataLine++;
            textBoxInfo.AppendText(String.Format("{0:00}", dataLine)
                      + " Write to " + doSubdev.deviceName
                      + ",  Device: " + doSubdev.deviceIndex.ToString()
                      + ",  Subdevice: " + doSubdev.subdevIndex.ToString()
                      + ",  Value hex: " + String.Format("0x{0:X}", valueToWrite)
                      + ",  bin: " + IntToBin(valueToWrite, doSubdev.numOfChannels)
                      + ",  dec: " + valueToWrite.ToString() + Environment.NewLine);
        }

        private int WriteSingleValue(DoSubdevice doSubdev, int valueToWrite)
        {
            meIDS.meIOSingle_t[] io_single = new meIDS.meIOSingle_t[1];     
            io_single[0].iDevice = doSubdev.deviceIndex;                    
            io_single[0].iSubdevice = doSubdev.subdevIndex;                 
            io_single[0].iChannel = 0;                                      
            io_single[0].iDir = meIDS.ME_DIR_OUTPUT;                        
            io_single[0].iValue = valueToWrite;                             
            io_single[0].iTimeOut = 0;                                      
            io_single[0].iFlags = doSubdev.singleTypeFlag;                  
            io_single[0].iErrno = 0;                                        

            int meError = meIDS.meIOSingle(ref io_single[0],                   
                                       1,                                   
                                       meIDS.ME_IO_SINGLE_NO_FLAGS);        

            return meError;
        }

        private int ConfigureSubdeviceForOutput(DoSubdevice doSubdev)
        {
            int meError;
            if ((doSubdev.deviceCaps & meIDS.ME_CAPS_DIO_SINK_SOURCE) == meIDS.ME_CAPS_DIO_SINK_SOURCE)
            {
                int singleConfig = ((ComboBoxItem)(coBoxPortConfiguration.SelectedItem)).GetValue;
                meError = meIDS.meIOSingleConfig(doSubdev.deviceIndex,          
                                                 doSubdev.subdevIndex,          
                                                 0,                             
                                                 singleConfig,                  
                                                 meIDS.ME_REF_NONE,             
                                                 meIDS.ME_VALUE_NOT_USED,       
                                                 meIDS.ME_VALUE_NOT_USED,       
                                                 meIDS.ME_VALUE_NOT_USED,       
                                                 doSubdev.singleConfigFlag);    
            }
            else
            {
                meError = meIDS.meIOSingleConfig(doSubdev.deviceIndex,          
                                                 doSubdev.subdevIndex,          
                                                 0,                             
                                                 meIDS.ME_SINGLE_CONFIG_DIO_OUTPUT, 
                                                 meIDS.ME_REF_NONE,             
                                                 meIDS.ME_TRIG_CHAN_DEFAULT,    
                                                 meIDS.ME_TRIG_TYPE_NONE,       
                                                 meIDS.ME_VALUE_NOT_USED,       
                                                 doSubdev.singleConfigFlag);    
            }
            return meError;
        }

        private void btnResetSubdev_Click(object sender, EventArgs e)
        {
            if (coBoxSubDevs.SelectedIndex >= 0)
            {
                DoSubdevice doSubdev;
                doSubdev = subdeviceList[coBoxSubDevs.SelectedIndex];
                int meError = meIDS.meIOResetSubdevice(doSubdev.deviceIndex,
                                                      doSubdev.subdevIndex,
                                                      meIDS.ME_IO_RESET_SUBDEVICE_NO_FLAGS);
                if (meError == meIDS.ME_ERRNO_SUCCESS)
                {
                    textBoxInfo.AppendText("Reset Device " + doSubdev.deviceIndex.ToString() + ", Subdevice " + doSubdev.subdevIndex.ToString() + Environment.NewLine);
                }
                else
                {
                    DisplayErrorMessage("Error resetting Subdevice", meError);
                }
            }
        }

        private void CloseMEiDSDriver()
        {
            if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS)
            {
                meIDS.meClose(meIDS.ME_CLOSE_NO_FLAGS);                         
                openErrSuccess = Constants.ME_IDS_NOTOPENED;
                subdeviceList.Clear();
                coBoxSubDevs.Items.Clear();
                btnResetSubdev.Enabled = false;
                btnSingleWrite.Enabled = false;
                coBoxPortConfiguration.Enabled = false;
            }
        }

        private int GetDevicePlugged(int idxDevice, out int plugged)
        {
            int vendorId;
            int deviceId;
            int serialNo;
            int busType;
            int busNo;
            int devNo;
            int funcNo;

            int meError = meIDS.meQueryInfoDevice(idxDevice,
                                             out vendorId,
                                             out deviceId,
                                             out serialNo,
                                             out busType,
                                             out busNo,
                                             out devNo,
                                             out funcNo,
                                             out plugged);              //Indicates whether a device is physically available
            return meError;
        }

        private void FillPortConfiguration()
        {
            coBoxPortConfiguration.Items.Add(new ComboBoxItem("High impedance", meIDS.ME_SINGLE_CONFIG_DIO_HIGH_IMPEDANCE));
            coBoxPortConfiguration.Items.Add(new ComboBoxItem("Sink", meIDS.ME_SINGLE_CONFIG_DIO_SINK));
            coBoxPortConfiguration.Items.Add(new ComboBoxItem("Source", meIDS.ME_SINGLE_CONFIG_DIO_SOURCE));
            coBoxPortConfiguration.SelectedIndex = 0;
        }

        private string GetSubDevTypeName(int subdeviceType)
        {
            string tmpType;
            switch (subdeviceType)
            {
                case meIDS.ME_TYPE_INVALID:
                    tmpType = "ME_TYPE_INVALID";
                    break;
                case meIDS.ME_TYPE_AI:
                    tmpType = "ME_TYPE_AI";
                    break;
                case meIDS.ME_TYPE_AO:
                    tmpType = "ME_TYPE_AO";
                    break;
                case meIDS.ME_TYPE_DIO:
                    tmpType = "ME_TYPE_DIO";
                    break;
                case meIDS.ME_TYPE_DO:
                    tmpType = "ME_TYPE_DO";
                    break;
                case meIDS.ME_TYPE_DI:
                    tmpType = "ME_TYPE_DI";
                    break;
                case meIDS.ME_TYPE_CTR:
                    tmpType = "ME_TYPE_CTR";
                    break;
                case meIDS.ME_TYPE_EXT_IRQ:
                    tmpType = "ME_TYPE_EXT_IRQ";
                    break;
                case meIDS.ME_TYPE_FIO:
                    tmpType = "ME_TYPE_FIO";
                    break;
                case meIDS.ME_TYPE_FO:
                    tmpType = "ME_TYPE_FO";
                    break;
                case meIDS.ME_TYPE_FI:
                    tmpType = "ME_TYPE_FI";
                    break;
                case meIDS.ME_TYPE_FPGA:
                    tmpType = "ME_TYPE_FPGA";
                    break;
                default:
                    tmpType = String.Format("0x{0:X}", subdeviceType);
                    break;
            };
            return tmpType;
        }

        private void DisplayVersionMainDriver()
        {
            int tmpVersion;
            meIDS.meQueryVersionMainDriver(out tmpVersion);
            textBoxInfo.AppendText("Maindriver version: " + MeVersionToString(tmpVersion) + Environment.NewLine);
        }

        private void DisplayVersionLibrary()
        {
            int tmpVersion;
            meIDS.meQueryVersionLibrary(out tmpVersion);
            textBoxInfo.AppendText("Library version: " + MeVersionToString(tmpVersion) + Environment.NewLine);
        }

        private void DisplayErrorMessage(string message, int meError)
        {
            StringBuilder errMessage = new StringBuilder(meIDS.ME_ERROR_MSG_MAX_COUNT);
            textBoxInfo.AppendText(message + meError.ToString() + Environment.NewLine);
            meIDS.meErrorGetMessage(meError, errMessage, errMessage.Capacity);             
            textBoxInfo.AppendText("ErrorMessage: " + errMessage + Environment.NewLine);   
        }

        private string MeVersionToString(int version)
        {
            return String.Format("{0}.{1}.{2}.{3}", version >> 24, 0xFF & (version >> 16), 0xFF & (version >> 8), 0xFF & version);
        }

        private string IntToBin(int val, int dispDigits)
        {
            int b;
            bool found = false;
            string tmpStr = "";
            for (b = 0; b < Constants.DIGITNUMS.Length; b++)
            {
                if (dispDigits == Constants.DIGITNUMS[b])
                {
                    found = true;
                    break;
                }
            }
            if (found)
            {
                if (dispDigits == 1)
                {
                    tmpStr = (val & 1).ToString();
                }
                else
                {
                    for (int i = (dispDigits >> 2) - 1; i >= 0; i--)
                    {
                        b = (val >> (i << 2)) & 0x0F;
                        tmpStr += Constants.BIN_STR[b];
                    }
                }
            }
            return tmpStr;
        }

        private void btnClearHwInfo_Click(object sender, EventArgs e)
        {
            textBoxInfo.Clear();
            dataLine = 0;
        }

        private void btnAbout_Click(object sender, EventArgs e)
        {
            AboutBox1 aboutBox = new AboutBox1();
            aboutBox.ShowDialog();
        }
    }

    class DoSubdevice
    {
        public string deviceName;
        public int deviceIndex;
        public int subdevIndex;
        public int numOfChannels;
        public int deviceCaps;
        public int singleConfigFlag;
        public int singleTypeFlag;
        
        public DoSubdevice(string DeviceName, int DeviceIndex, int SubdevIndex, int NumOfChannels, int DeviceCaps)
        {
            this.deviceName = DeviceName;
            this.deviceIndex = DeviceIndex;
            this.subdevIndex = SubdevIndex;
            this.numOfChannels = NumOfChannels;
            this.deviceCaps = DeviceCaps;

            switch (NumOfChannels)
            {
                case 1:
                    singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_BIT;           
                    singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_BIT;
                    break;
                case 8:
                    singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_BYTE;          
                    singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_BYTE;
                    break;
                case 16:
                    singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_WORD;         
                    singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_WORD;
                    break;
                case 24:
                    singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_DWORD;        
                    singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_DWORD;
                    break;
                case 32:
                    singleConfigFlag = meIDS.ME_IO_SINGLE_CONFIG_DIO_DWORD;        
                    singleTypeFlag = meIDS.ME_IO_SINGLE_TYPE_DIO_DWORD;
                    break;
                default:
                    singleConfigFlag = meIDS.ME_VALUE_INVALID;
                    singleTypeFlag = meIDS.ME_VALUE_INVALID;
                    break;
            }
        }
    }

    public class ComboBoxItem
    {
        string dispValue;
        int val;

        public ComboBoxItem(string displayValue, int value)
        {
            dispValue = displayValue;
            val = value;
        }

        public override string ToString()
        {
            return dispValue;
        }

        public int GetValue
        {
            get
            {
                return val;
            }
        }
    }

    static class Constants
    {
        public const int ME_IDS_NOTOPENED = -1;
        public static readonly string[] BIN_STR = {"0000", "0001", "0010", "0011",
                                                   "0100", "0101", "0110", "0111",
                                                   "1000", "1001", "1010", "1011",
                                                   "1100", "1101", "1110", "1111"};
        public static int[] DIGITNUMS = { 1, 8, 16, 24, 32 };
    }
}
