﻿//==============================================================================
//Copyright(c), Meilhaus Electronic GmbH
//              http://www.meilhaus.de
//              support@meilhaus.de
//
//Source File : FormMain.cs
//Description : Example program for writing to Frequency-out (FO) subdevices
//              - Open ME-iDS main driver when program is started
//              - Go through the list of all devices and find FO subdevices
//              - Add FO subdevices to the combobox
//              - Button "Set config": Set configuration of the selected subdevice or FO-channel
//                before data can be written
//              - Click on button "Write Data" to emit the signal depending on selected flags
//              - Written values can be read back by click on "Read Data"
//              - Close ME-iDS main driver when program is closed
//==============================================================================
//  To emit a digital signal two values of the FO subdevice must be set:
//  - 1. "Ticks total" = value of the entire period of the signal
//  - 2. "Ticks first phase" = value of the first phase of the signal (first phase can be high or low)
//    Therefore the function meIOSingle() must be called twice using different flags
//
//  digital signal:
//      +-------------------+         +---
//      |                   |         |
//      |                   |         |
//  ----+                   +---------+
//
//      ^-------Ticks total-----------^
//
//      ^--Ticks 1st phase--^
//
//
//  How to start several (a group of) FO-subdevices (FO-channels) synchronously:
//  - configure each subdevice with the flag 'ME_TRIG_CHAN_SYNCHRONOUS'
//  - write the values for 'Ticks total' and 'Ticks first phase' to each channel without the flag 'Synchronous',
//  - set the flag 'ME_TRIG_CHAN_SYNCHRONOUS' for the last subdevice in the group.
//  Port configuration:
//  - Configuration of the output driver of devices with sink, source, high impedance capabilities (e.g. ME-5810A, ME-5002)
//  Configuration:
//    Immediate output = start the output of the configured subdevice(s) or channel(s)
//    after values "Ticks total" and "Ticks first phase" are set
//    Synchronous output = don't start the selected subdevice or channel, set only values for "Ticks total" and "Ticks first phase"
//  - (Also see ME-iDS Help section: Supported Hardware / ME-46xx Description / Special features / Analog output subdevices)
//  Read:
//  - read back the values of the previously emitted signal
//    Ticks total: the value of the entire period
//    Ticks first phase: the value of the first phase of the period
//  Write:
//  - Signal start level: (Start signal with high or low phase)
//    High: first phase of the signal will be high
//    Low: first phase of the signal will he low
//  - Signal start type: (start immediately when button clicked or at the end of the currently emitted signal period
//    Hard: start the output of the signal immediately (on button click)
//    Soft: Controlled start of the new signal when the period of the current signal (if any) is over
//  - Write mode:
//    Direct: This subdevice or channel will be started alone
//    Synchronous: All those subdevices which have been configured with the 'iTrigChan' = ME_TRIG_CHAN_SYNCHRONOUS
//    in meIOSingleConfig above will be started simultaneously together with this subdevice
//------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using meIDS_CS;

namespace Single_Write_Frequency
{
    public partial class FormMain : Form
    {
        private int openErrSuccess = Constants.ME_IDS_NOTOPENED;
        private int lineNumber = 0;
        private List<FoSubdevice> subdeviceList = new List<FoSubdevice>();

        public FormMain()
        {
            InitializeComponent();
        }

        private void FormMain_Shown(object sender, EventArgs e)
        {
            txtBoxFirstPhase.Text = String.Format("{0:G}", 0.2);
            FillPortConfiguration();
            OpenMEiDSDriver();
        }

        private void FormMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            CloseMEiDSDriver();
        }

        private void btnClose_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void OpenMEiDSDriver()
        {
            try
            {
                textBoxInfo.AppendText("Opening Maindriver..." + Environment.NewLine);
                openErrSuccess = meIDS.meOpen(meIDS.ME_OPEN_NO_FLAGS);

                if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS)
                {
                    DisplayVersionLibrary();
                    DisplayVersionMainDriver();
                    textBoxInfo.AppendText(Environment.NewLine);

                    int numDevices;
                    meIDS.meQueryNumberDevices(out numDevices);
                    textBoxInfo.AppendText("Number of devices found: " + numDevices.ToString() + Environment.NewLine);

                    GetFoSubdevices(numDevices);
                }
                else
                {
                    DisplayErrorMessage("meOpen - Error: ", openErrSuccess);
                }
            }
            catch (DllNotFoundException ex)
            {
                textBoxInfo.AppendText("The ME-iDS is not properly installed." + Environment.NewLine);
                textBoxInfo.AppendText(ex.Message + Environment.NewLine);
            }
        }

        private void GetFoSubdevices(int numDevices)
        {
            if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS && numDevices > 0)
            {
                textBoxInfo.AppendText("Retrieving subdevices of type ME_TYPE_FO (Frequency-Out)..." + Environment.NewLine);

                for (int idxDevice = 0; idxDevice < numDevices; idxDevice++)
                {
                    int plugged;
                    int meError = GetDevicePlugged(idxDevice, out plugged);

                    if (meError == meIDS.ME_ERRNO_SUCCESS)
                    {
                        if (plugged == meIDS.ME_PLUGGED_IN)
                        {
                            GetSubdevicesByType(idxDevice, meIDS.ME_TYPE_FO);
                        }
                        else
                        {
                            textBoxInfo.AppendText("Device " + idxDevice.ToString() + " not plugged in" + Environment.NewLine);
                        }
                    }
                }
            }
            DisplaySubdevices();
        }

        private void GetSubdevicesByType(int idxDevice, int subdeviceType)
        {
            int numSubdevices;
            int idxMatchedSubdevice;
            StringBuilder devName = new StringBuilder(meIDS.ME_DEVICE_NAME_MAX_COUNT);

            int meError = meIDS.meQueryNumberSubdevices(idxDevice, out numSubdevices);
            if (meError == meIDS.ME_ERRNO_SUCCESS)
            {
                meIDS.meQueryNameDevice(idxDevice, devName, devName.Capacity);

                int idxSubdevice = 0;
                while (idxSubdevice < numSubdevices)
                {
                    meError = meIDS.meQuerySubdeviceByType(idxDevice,
                                                           idxSubdevice,
                                                           subdeviceType,
                                                           meIDS.ME_SUBTYPE_ANY,
                                                           out idxMatchedSubdevice);

                    if (meError == meIDS.ME_ERRNO_SUCCESS)
                    {
                        AddFoSubdevice(devName.ToString(), idxDevice, idxMatchedSubdevice, subdeviceType);
                        idxSubdevice = idxMatchedSubdevice + 1;
                    }
                    else
                    {
                        break;
                    }
                }
            }
        }

        private void AddFoSubdevice(String devName, int idxDevice, int idxSubdevice, int subdeviceType)
        {
            int numOfChannels;
            int devCaps;

            meIDS.meQueryNumberChannels(idxDevice, idxSubdevice, out numOfChannels);
            meIDS.meQuerySubdeviceCaps(idxDevice, idxSubdevice, out devCaps);
            subdeviceList.Add(new FoSubdevice(devName, idxDevice, idxSubdevice, numOfChannels, devCaps));

            coBoxSubDevs.Items.Add(devName
                                    + ",  Device " + idxDevice.ToString()
                                    + ",  Subdevice " + idxSubdevice.ToString()
                                    + ",  Type: " + GetSubDevTypeName(subdeviceType)
                                    + ",  Channels: " + numOfChannels.ToString());
        }

        private void btnResetSubdev_Click(object sender, EventArgs e)
        {
            FoSubdevice foSubdev;
            foSubdev = subdeviceList[coBoxSubDevs.SelectedIndex];
            int meError = meIDS.meIOResetSubdevice(foSubdev.deviceIndex,
                                                  foSubdev.subdevIndex,
                                                  meIDS.ME_IO_RESET_SUBDEVICE_NO_FLAGS);
            if (meError == meIDS.ME_ERRNO_SUCCESS)
            {
                textBoxInfo.AppendText("Reset Device " + foSubdev.deviceIndex.ToString() + ", Subdevice " + foSubdev.subdevIndex.ToString() + Environment.NewLine);
            }
            else
            {
                DisplayErrorMessage("Error resetting Subdevice", meError);
            }
        }

        private void btnSetConfig_Click(object sender, EventArgs e)
        {
            if (coBoxSubDevs.SelectedIndex >= 0)
            {
                FoSubdevice foSubDevice = subdeviceList[coBoxSubDevs.SelectedIndex];

                int meError = ConfigureFoSubdevice(foSubDevice, coBoxChannel.SelectedIndex);
                if (meError == meIDS.ME_ERRNO_SUCCESS)
                {
                    textBoxInfo.AppendText("Set config: Device " + foSubDevice.deviceIndex.ToString()
                                     + ", Subdevice " + foSubDevice.subdevIndex.ToString()
                                     + ", Channel " + coBoxChannel.SelectedIndex.ToString()
                                     + ", " + GetFoConfigText()
                                     + Environment.NewLine);
                }
                else
                {
                    DisplayErrorMessage("meIOSingleConfig - Error: ", meError);
                }
            }
        }

        private int ConfigureFoSubdevice(FoSubdevice foSubdevice, int idxChannel)
        {
            int singleConfig = GetSingleConfigParam(foSubdevice);
            int triggerChannel = GetOutputParam();
            int meError = meIDS.meIOSingleConfig(foSubdevice.deviceIndex,
                                        foSubdevice.subdevIndex,
                                        idxChannel,
                                        singleConfig,
                                        meIDS.ME_VALUE_NOT_USED,
                                        triggerChannel,
                                        meIDS.ME_TRIG_TYPE_NONE,
                                        meIDS.ME_VALUE_NOT_USED,
                                        meIDS.ME_IO_SINGLE_CONFIG_NO_FLAGS);

            return meError;
        }

        private int GetSingleConfigParam(FoSubdevice foSubdevice)
        {
            int singleConfig;
            if ((foSubdevice.subdeviceCaps & meIDS.ME_CAPS_FIO_SINK_SOURCE) == meIDS.ME_CAPS_FIO_SINK_SOURCE)
            {
                singleConfig = ((ComboBoxItem)(coBoxPortConfiguration.SelectedItem)).GetValue;
            }
            else
            {
                singleConfig = meIDS.ME_SINGLE_CONFIG_FIO_OUTPUT;
            }
            return singleConfig;
        }

        private int GetOutputParam()
        {
            int triggerChannel;

            if (rBtnImmediateOutput.Checked)
            {
                triggerChannel = meIDS.ME_TRIG_CHAN_DEFAULT;
            }
            else
            {
                triggerChannel = meIDS.ME_TRIG_CHAN_SYNCHRONOUS;
            }
            return triggerChannel;
        }

        private string GetFoConfigText()
        {
            string trigChan;
            if (rBtnImmediateOutput.Checked)
            {
                trigChan = rBtnImmediateOutput.Text;
            }
            else
            {
                trigChan = rBtnSynchronousOutput.Text;
            }
            return trigChan;
        }

        private void btnReadData_Click(object sender, EventArgs e)
        {
            if (coBoxSubDevs.Items.Count > 0)
            {
                FoSubdevice  foSubdevice = subdeviceList[coBoxSubDevs.SelectedIndex];
                int idxChannel = coBoxChannel.SelectedIndex;
                textBoxInfo.AppendText(Environment.NewLine + "Read Data" + Environment.NewLine);

                int valueTotal;
                int meError = ReadSingleValue(foSubdevice, idxChannel, meIDS.ME_IO_SINGLE_TYPE_FIO_TICKS_TOTAL, out valueTotal);

                if (meError == meIDS.ME_ERRNO_SUCCESS)
                {
                    int value1stPhase;
                    meError = ReadSingleValue(foSubdevice, idxChannel, meIDS.ME_IO_SINGLE_TYPE_FIO_TICKS_FIRST_PHASE, out value1stPhase);

                    if (meError == meIDS.ME_ERRNO_SUCCESS)
                    {
                        DisplayReadBackValues(foSubdevice, idxChannel, value1stPhase, valueTotal);
                    }
                }
                if (meError != meIDS.ME_ERRNO_SUCCESS)
                {
                    DisplayErrorMessage("meIOSingle - Error: ", meError);
                }
            }
        }

        private int ReadSingleValue(FoSubdevice foSubdevice, int idxChannel, int flags, out int value)
        {
            meIDS.meIOSingle_t[] io_single = new meIDS.meIOSingle_t[1];

            io_single[0].iDevice = foSubdevice.deviceIndex;
            io_single[0].iSubdevice = foSubdevice.subdevIndex;
            io_single[0].iChannel = idxChannel;
            io_single[0].iDir = meIDS.ME_DIR_INPUT;
            io_single[0].iValue = 0;
            io_single[0].iTimeOut = meIDS.ME_VALUE_NOT_USED;
            io_single[0].iFlags = flags;

            int meError = meIDS.meIOSingle(ref io_single[0],
                       1,
                       meIDS.ME_IO_SINGLE_NO_FLAGS);

            value = io_single[0].iValue;
            return meError;
        }

        private void DisplayReadBackValues(FoSubdevice foSubdevice, int idxChannel, int value1stPhase, int valueTotal)
        {
            lineNumber++;
            textBoxInfo.AppendText(String.Format("{0:00}", lineNumber)
                                + " Read from " + foSubdevice.deviceName
                                + ",  Device: " + foSubdevice.deviceIndex.ToString()
                                + ",  Subdevice: " + foSubdevice.subdevIndex.ToString()
                                + ",  Channel: " + idxChannel.ToString()
                                + Environment.NewLine);
            textBoxInfo.AppendText("  Ticks total: " + valueTotal.ToString()
                                + Environment.NewLine);
            textBoxInfo.AppendText("  Ticks first phase: " + value1stPhase.ToString()
                                + Environment.NewLine);
        }

        private void btnWriteData_Click(object sender, EventArgs e)
        {
            if (coBoxSubDevs.SelectedIndex >= 0)
            {
                FoSubdevice foSubdevice = subdeviceList[coBoxSubDevs.SelectedIndex];
                int idxChannel = coBoxChannel.SelectedIndex;
                textBoxInfo.AppendText(Environment.NewLine + "Write Data" + Environment.NewLine);

                try
                {
                    Double secondsAchievedFirstPhase = double.Parse(txtBoxFirstPhase.Text);
                    Double secondsAchivedTotal = double.Parse(txtBoxTotalTime.Text);

                    int ticksTotal;
                    int meError = ConvertTimeToTicks(foSubdevice, meIDS.ME_TIMER_FIO_TOTAL, ref secondsAchivedTotal, out ticksTotal);
                    if (meError == meIDS.ME_ERRNO_SUCCESS)
                    {
                        int flags = GetFoTotalTimeFlags();
                        meError = WriteTicks(foSubdevice, idxChannel, flags, ticksTotal);

                        if (meError == meIDS.ME_ERRNO_SUCCESS)
                        {
                            int ticksFirstPhase;
                            meError = ConvertTimeToTicks(foSubdevice, meIDS.ME_TIMER_FIO_FIRST_PHASE, ref secondsAchievedFirstPhase, out ticksFirstPhase);
                            if (meError == meIDS.ME_ERRNO_SUCCESS)
                            {
                                flags = GetFoFirstPhaseTimeFlags();
                                meError = WriteTicks(foSubdevice, idxChannel, flags, ticksFirstPhase);
                                if (meError == meIDS.ME_ERRNO_SUCCESS)
                                {
                                    DisplayTimeValues(foSubdevice, idxChannel, secondsAchivedTotal, secondsAchievedFirstPhase, ticksTotal, ticksFirstPhase);
                                }
                            }
                        }
                    }
                    if(meError != meIDS.ME_ERRNO_SUCCESS)
                    {
                        DisplayErrorMessage("Error: ", meError);
                    }
                }
                catch (System.Exception ex)
                {
                    textBoxInfo.AppendText("Error: " + ex.Message);
                }
            }
        }

        private void DisplayTimeValues(FoSubdevice foSubdevice, int idxChannel, Double secondsTotal, Double secondsFirstPhase, int ticksTotal, int ticksFirstPhase)
        {
            lineNumber++;
            textBoxInfo.AppendText(String.Format("{0:00}", lineNumber)
                               + " Write to " + foSubdevice.deviceName
                               + ",  Device: " + foSubdevice.deviceIndex.ToString()
                               + ",  Subdevice: " + foSubdevice.subdevIndex.ToString()
                               + ",  Channel: " + idxChannel.ToString()
                               + Environment.NewLine);

            textBoxInfo.AppendText(String.Format("  Total time: {0:0.##########} sec, frequency: {1:0.##########} Hz", secondsTotal, 1 / secondsTotal)
                               + ",  Ticks: " + ticksTotal.ToString()
                               + Environment.NewLine);

            textBoxInfo.AppendText(String.Format("  First phase time: {0:0.##########} sec", secondsFirstPhase)
                               + ",  Ticks: " + ticksFirstPhase.ToString()
                               + Environment.NewLine);

        }

        private int GetFoFirstPhaseTimeFlags()
        {
            int flags = meIDS.ME_IO_SINGLE_TYPE_FIO_TICKS_FIRST_PHASE;
            if (rBtnSynchronousOutput.Checked)
            {
                flags |= meIDS.ME_IO_SINGLE_TYPE_TRIG_SYNCHRONOUS;
            }
            if (rBtnStartLevelLow.Checked)
            {
                flags |= meIDS.ME_IO_SINGLE_TYPE_FO_START_LOW;
            }

            if (rBtnStarttypeSoft.Checked)
            {
                flags |= meIDS.ME_IO_SINGLE_TYPE_FO_START_SOFT;
            }
            return flags;
        }

        private int GetFoTotalTimeFlags()
        {
            int flags = meIDS.ME_IO_SINGLE_TYPE_NO_FLAGS;
            if (rBtnSynchronousOutput.Checked)
            {
                flags |= meIDS.ME_IO_SINGLE_TYPE_TRIG_SYNCHRONOUS;
            }
            flags |= meIDS.ME_IO_SINGLE_TYPE_FIO_TICKS_TOTAL
                  | meIDS.ME_IO_SINGLE_TYPE_FO_UPDATE_ONLY;
            return flags;
        }

        private int ConvertTimeToTicks(FoSubdevice foSubdevice, int timer, ref Double secondsAchived, out int ticks)
        {
            int ticksHigh;
            int meError = meIDS.meIOSingleTimeToTicks(foSubdevice.deviceIndex,
                    foSubdevice.subdevIndex,
                    timer,
                    ref secondsAchived,
                    out ticks,
                    out ticksHigh,
                    meIDS.ME_IO_SINGLE_TIME_TO_TICKS_NO_FLAGS);

            return meError;
        }

        private int WriteTicks(FoSubdevice foSubdevice, int idxChannel, int flags, int ticks)
        {
            meIDS.meIOSingle_t[] io_single = new meIDS.meIOSingle_t[1];

            io_single[0].iDevice = foSubdevice.deviceIndex;
            io_single[0].iSubdevice = foSubdevice.subdevIndex;
            io_single[0].iChannel = idxChannel;
            io_single[0].iDir = meIDS.ME_DIR_OUTPUT;
            io_single[0].iValue = ticks;
            io_single[0].iTimeOut = 0;
            io_single[0].iFlags = flags;

            int meError = meIDS.meIOSingle(ref io_single[0],
                                           1,
                                           meIDS.ME_IO_SINGLE_NO_FLAGS);
            return meError;
        }

        private void CloseMEiDSDriver()
        {
            if (openErrSuccess == meIDS.ME_ERRNO_SUCCESS)
            {
                meIDS.meClose(meIDS.ME_CLOSE_NO_FLAGS);
                openErrSuccess = Constants.ME_IDS_NOTOPENED;
                btnReadData.Enabled = false;
                btnResetSubdev.Enabled = false;
                btnSetConfig.Enabled = false;
                subdeviceList.Clear();
                coBoxSubDevs.Items.Clear();
                coBoxChannel.Items.Clear();
            }
        }

        private void coBoxSubDevs_SelectedIndexChanged(object sender, EventArgs e)
        {
            coBoxChannel.Items.Clear();
            if (coBoxSubDevs.SelectedIndex >= 0)
            {
                FoSubdevice foSubdevice;
                foSubdevice = subdeviceList[coBoxSubDevs.SelectedIndex];
                coBoxPortConfiguration.Enabled = ((foSubdevice.subdeviceCaps & meIDS.ME_CAPS_FIO_SINK_SOURCE) == meIDS.ME_CAPS_FIO_SINK_SOURCE);
                FillChannels(foSubdevice);
            }
        }

        private void FillChannels(FoSubdevice foSubdevice)
        {
            for (int i = 0; i < foSubdevice.numOfChannels; i++)
            {
                coBoxChannel.Items.Add(i.ToString());
            }
            if (coBoxChannel.Items.Count > 0)
                coBoxChannel.SelectedIndex = 0;
        }

        private void DisplaySubdevices()
        {
            if (coBoxSubDevs.Items.Count > 0)
            {
                textBoxInfo.AppendText("Number of Frequency-Out subdevices found: " + coBoxSubDevs.Items.Count.ToString() + Environment.NewLine);
                for (int i = 0; i < coBoxSubDevs.Items.Count; i++)
                {
                    textBoxInfo.AppendText(coBoxSubDevs.Items[i] + Environment.NewLine);
                }
                textBoxInfo.AppendText(Environment.NewLine);
                coBoxSubDevs.SelectedIndex = 0;
                btnReadData.Enabled = true;
                btnResetSubdev.Enabled = true;
                btnWriteData.Enabled = true;
                btnSetConfig.Enabled = true;
            }
            else
            {
                textBoxInfo.AppendText("No Frequency-Out subdevices found." + Environment.NewLine);
                CloseMEiDSDriver();
            }
        }

        private void DisplayErrorMessage(string message, int meError)
        {
            StringBuilder errMessage = new StringBuilder(meIDS.ME_ERROR_MSG_MAX_COUNT);
            textBoxInfo.AppendText(message + meError.ToString() + Environment.NewLine);
            meIDS.meErrorGetMessage(meError, errMessage, errMessage.Capacity);
            textBoxInfo.AppendText("ErrorMessage: " + errMessage + Environment.NewLine);
        }

        private void FillPortConfiguration()
        {
            coBoxPortConfiguration.Items.Add(new ComboBoxItem("High impedance", meIDS.ME_SINGLE_CONFIG_FIO_HIGH_IMPEDANCE));
            coBoxPortConfiguration.Items.Add(new ComboBoxItem("Sink", meIDS.ME_SINGLE_CONFIG_FIO_SINK));
            coBoxPortConfiguration.Items.Add(new ComboBoxItem("Source", meIDS.ME_SINGLE_CONFIG_FIO_SOURCE));
            coBoxPortConfiguration.SelectedIndex = 0;
        }

        private void DisplayVersionMainDriver()
        {
            int tmpVersion;
            meIDS.meQueryVersionMainDriver(out tmpVersion);
            textBoxInfo.AppendText("Maindriver version: " + MeVersionToString(tmpVersion) + Environment.NewLine);
        }

        private int DisplayVersionLibrary()
        {
            int tmpVersion;
            meIDS.meQueryVersionLibrary(out tmpVersion);
            textBoxInfo.AppendText("Library version: " + MeVersionToString(tmpVersion) + Environment.NewLine);
            return tmpVersion;
        }

        private int GetDevicePlugged(int idxDevice, out int plugged)
        {
            int vendorId;
            int deviceId;
            int serialNo;
            int busType;
            int busNo;
            int devNo;
            int funcNo;

            int meError = meIDS.meQueryInfoDevice(idxDevice,
                                             out vendorId,
                                             out deviceId,
                                             out serialNo,
                                             out busType,
                                             out busNo,
                                             out devNo,
                                             out funcNo,
                                             out plugged);
            return meError;
        }

        private string GetSubDevTypeName(int subdeviceType)
        {
            string tmpType;
            switch (subdeviceType)
            {
                case meIDS.ME_TYPE_INVALID:
                    tmpType = "ME_TYPE_INVALID";
                    break;
                case meIDS.ME_TYPE_AI:
                    tmpType = "ME_TYPE_AI";
                    break;
                case meIDS.ME_TYPE_AO:
                    tmpType = "ME_TYPE_AO";
                    break;
                case meIDS.ME_TYPE_DIO:
                    tmpType = "ME_TYPE_DIO";
                    break;
                case meIDS.ME_TYPE_DO:
                    tmpType = "ME_TYPE_DO";
                    break;
                case meIDS.ME_TYPE_DI:
                    tmpType = "ME_TYPE_DI";
                    break;
                case meIDS.ME_TYPE_CTR:
                    tmpType = "ME_TYPE_CTR";
                    break;
                case meIDS.ME_TYPE_EXT_IRQ:
                    tmpType = "ME_TYPE_EXT_IRQ";
                    break;
                case meIDS.ME_TYPE_FIO:
                    tmpType = "ME_TYPE_FIO";
                    break;
                case meIDS.ME_TYPE_FO:
                    tmpType = "ME_TYPE_FO";
                    break;
                case meIDS.ME_TYPE_FI:
                    tmpType = "ME_TYPE_FI";
                    break;
                case meIDS.ME_TYPE_FPGA:
                    tmpType = "ME_TYPE_FPGA";
                    break;
                default:
                    tmpType = String.Format("0x{0:X}", subdeviceType);
                    break;
            };
            return tmpType;
        }

        private void btnClearHwInfo_Click(object sender, EventArgs e)
        {
            textBoxInfo.Clear();
            lineNumber = 0;
        }

        private string MeVersionToString(int version)
        {
            return String.Format("{0}.{1}.{2}.{3}", version >> 24, 0xFF & (version >> 16), 0xFF & (version >> 8), 0xFF & version);
        }

        private void btnAbout_Click(object sender, EventArgs e)
        {
            AboutBox1 aboutBox = new AboutBox1();
            aboutBox.ShowDialog();
        }
    }

    class FoSubdevice
    {
        public string deviceName;
        public int deviceIndex;
        public int subdevIndex;
        public int numOfChannels;
        public int subdeviceCaps;
        public FoSubdevice(string DeviceName, int DeviceIndex, int SubdevIndex, int NumOfChannels, int SubdeviceCaps)
        {
            this.deviceName = DeviceName;
            this.deviceIndex = DeviceIndex;
            this.subdevIndex = SubdevIndex;
            this.numOfChannels = NumOfChannels;
            this.subdeviceCaps = SubdeviceCaps;
        }
    }

    public class ComboBoxItem
    {
        string dispValue;
        int val;

        public ComboBoxItem(string displayValue, int value)
        {
            dispValue = displayValue;
            val = value;
        }

        public override string ToString()
        {
            return dispValue;
        }

        public int GetValue
        {
            get
            {
                return val;
            }
        }
    }

    static class Constants
    {
        public const int ME_IDS_NOTOPENED = -1;
    }
}
